/* 
   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifndef __DHCP_H__
#define __DHCP_H__

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

#define GUID_LEN 16
/* Define a DHCP version of the bootp packet for nameing
   consistantely. This one is slightly different till we remove the old bootp
   specific version, as this one is more tested with various DHCP servers. *./
/* standard DHCP vendor field size */
//#ifdef NILO_PXE_COMPLY
#define DHCP_VEND_LEN  312              /* Intel defines lots of new options */
//#else
//#define DHCP_VEND_LEN  64               /* This is the RFC1533 length */
//#endif

#define DHCP_TIMER_FREQ 100

#define DHCP_CHADDR_LEN	16
#define DHCP_SNAME_LEN	64
#define DHCP_FILE_LEN	128
#define DHCP_MINPKTSZ	300	/* to check sizeof(dhcp_t) */

#define TAG_END			((u_int8_t) 255)
#define TAG_PAD			((u_int8_t)   0)

struct dhcp
{
    u_int8_t bp_opcode;		        /* packet opcode type */
    u_int8_t bp_htype;		        /* hardware addr type */
    u_int8_t bp_hlen;		        /* hardware addr length */
    u_int8_t bp_hops;		        /* gateway hops */
    u_int32_t bp_xid;		        /* transaction ID */
    u_int16_t bp_secs;		        /* seconds since boot began */
    u_int16_t bp_flags;		        /* RFC1532 broadcast, etc. */
    ip4_t bp_ciaddr;		        /* client IP address */
    ip4_t bp_yiaddr;		        /* your IP address */
    ip4_t bp_siaddr;		        /* server IP address */
    ip4_t bp_giaddr;		        /* gateway IP address */
    u_int8_t bp_chaddr[DHCP_CHADDR_LEN];  /* client hardware address */
    char bp_sname[DHCP_SNAME_LEN];	/* server host name */
    char bp_file[DHCP_FILE_LEN];	        /* boot file name */
    /* note that bp_vend can be longer, extending to end of packet. */
    u_int8_t bp_options[DHCP_VEND_LEN];      /* vendor-specific area */
};
typedef struct dhcp dhcp_t;

/* DHCP option and value (cf. RFC1533) */
typedef enum dhcp_opt {
    DHCP_PADOPTION =  0,                /* pad character */
    DHCP_ENDOPTION = 255,               /* option list terminator */
    DHCP_SUBNET_MASK = 1,                /* subnet mask */
    DHCP_TIMER_OFFSET = 2,              /* time offset */
    DHCP_ROUTERS_ON_SUBNET = 3,         /* router option */
    DHCP_TIME_SERVER = 4,               /* time server */
    DHCP_NAME_SERVER = 5,               /* name server */
    DHCP_DNS = 6,                       /* domain name server */
    DHCP_LOG_SERVER = 7,                /* log server */
    DHCP_COOKIE_SERVER = 8,             /* cookie server */
    DHCP_LPR_SERVER = 9,
    DHCP_IMPRES_SSERVER = 10,
    DHCP_RESOURCELOCATION_SERVER = 11,
    DHCP_HOSTNAME = 12,                 /* hostname */
    DHCP_BOOTFILESIZE = 13,             /* Boot File Size */
    DHCP_MERITDUMPFILE = 14,
    DHCP_DOMAINNAME = 15,               /* domainname */
    DHCP_SWAPSERVER = 16,
    DHCP_ROOTPATH = 17,
    DHCP_EXTENTIONS_PATH = 18,
    DHCP_IP_FORWARDING = 19,
    DHCP_NONLOCAL_SOURCE_ROUTING = 20,
    DHCP_POLICYFILTER = 21,
    DHCP_MAX_DGRAMREASM_SIZE = 22,
    DHCP_DEFAULT_IPTTL = 23,
    DHCP_PATH_MTU_AGING_TIMEOUT = 24,
    DHCP_PATH_MTU_PLATEAU_TABLE = 25,
    DHCP_IFMTU = 26,
    DHCP_ALLSUBNETS_LOCAL = 27,
    DHCP_BROADCAST_ADDR = 28,
    DHCP_PERFORM_MASK_DISCOVERY = 29,
    DHCP_MASK_SUPPLIER = 30,
    DHCP_PERFORM_ROUTER_DISCOVERY = 31,
    DHCP_ROUTER_SOLICITATION_ADDR = 32,
    DHCP_STATIC_ROUTE = 33,
    DHCP_TRAILER_ENCAPSULATION = 34,
    DHCP_ARPCACHE_TIMEOUT = 35,
    DHCP_ETHERNET_ENCAPSULATION = 36,
    DHCP_TCP_DEFAULT_TTL = 37,
    DHCP_TCP_KEEPALIVE_INTERVAL = 38,
    DHCP_TCP_KEEPALIVE_GARBAGE = 39,
    DHCP_NIS_DOMAINNAME = 40,
    DHCP_NIS_SERVERS = 41,
    DHCP_NTP_SERVERS = 42,
    DHCP_VENDOR_SPECIFICINFO = 43,      /* vendor specific */
    DHCP_NETBIOS_NAMESERVER = 44,       /* NetBIOS/WINS name server */
    DHCP_NETBIOS_DGRAMDISTSERVER = 45,  /* NetBIOS datagram distribution server */
    DHCP_NETBIOS_NODETYPE = 46,         /* NetBIOS node type */
    DHCP_NETBIOS_SCOPE = 47,            /* NetBIOS scope */
    DHCP_XFONT_SERVER = 48,
    DHCP_XDISPLAY_MANAGER = 49,
    DHCP_REQUESTED_IP_ADDR = 50,        /* requested IP address */
    DHCP_IPADDR_LEASE_TIME = 51,        /* IP addr lease time */
    DHCP_OPTION_OVERLOAD = 52,          /* option overload */
    DHCP_MESSAGE_TYPE = 53,             /* DHCP message type opcode */
    DHCP_SERVER_IDENTIFIER = 54,        /* server ID */
    DHCP_PARAM_REQUEST = 55,            /* parameter request list */
    DHCP_MSG = 56,                      /* message */
    DHCP_MAXMSGSIZE = 57,               /* maximum message size */
    DHCP_T1VALUE = 58,
    DHCP_T2VALUE = 59,
    DHCP_CLASS_IDENTIFIER = 60,         /* class identifier */
    DHCP_CLIENT_IDENTIFIER = 61,        /* client identifier */
    DHCP_BOOTFILE = 67,                 /* bootfile overload */
    DHCP_WWWSRV = 72,                   /* web server address */
    DHCP_SYSARCH = 93,                  /* system architecture */
    DHCP_NICIF = 94,                    /* NIC Interface specifier */
    DHCP_PLATFORMID = 97,               /* Platform ID */
    DHCP_PXE1_CMD_LINE = 135,           /* Bpbatch command line for PXE version 1.0 */
    DHCP_PXE_CMD_LINE = 155             /* Bpbatch command line for PXE version 2.0 */
} dhcp_opt_t;

/* DHCP option and value (cf. RFC1533) */
typedef enum dhcp_pxe_opts {
    DHCP_PXE_MTFTP_IP = 1,
    DHCP_PXE_MTFTP_CPORT = 2,
    DHCP_PXE_MTFTP_SPORT = 3,
    DHCP_PXE_MTFTP_TIMEOUT = 4,
    DHCP_PXE_MTFTP_DELAY = 5,
    DHCP_PXE_DISCOVERY_CNTRL = 6,
    DHCP_PXE_DISCOVERY_MCAST_ADDR = 7,
    DHCP_PXE_BOOT_SERVERS = 8,
    DHCP_PXE_BOOT_MENU = 9,
    DHCP_MENU_PROMPT = 10,
    DHCP_MCAST_ADDRS_ALLOC = 11,
    DHCP_PXE_CREDENTIALS = 12,
    DHCP_PXE_BOOT_ITEM = 71            /* PXE boot server type */
} dhcp_pxe_vendor_opts_t;

#define DHCP_MIN_OPT_SIZE 312
#define DHCP_MAX_OPT_SIZE DHCP_VEND_LEN /* was 576 */
#define DHCP_FILESPEC_SIZE 80
#define DHCP_ADDR_LIST_SIZE 8

#define VM_RFC1048      0x63825363L     /* RFC1048 magic number */

enum {
    DEFAULT_LEASE_TIME = 180*60,        /* default lease time in second */
    INFINITE_LEASE_TIME = 0xffffffff,   /* infinite lease time */
};

enum {
    N_REXMIT_DISCOVER = 4,              /* # of rexmit of DHCPDISCOVER msg */
    N_REXMIT_REQUEST = 10,
    N_REXMIT_REQ_REBOOT = 4,
};


/* UDP port numbers for TFTP */
typedef enum {
    DHCP_SERVER_PORT = 67,              /* from client to server */
    DHCP_CLIENT_PORT = 68               /* from server to client */
} dhcp_port_t;

/* DHCP message OP code */
typedef enum {
    DHCP_BOOTREQUEST = 1,
    DHCP_BOOTREPLY = 2
} dhcp_msg_opcode_t;

/* DHCP message type */
typedef enum {
    DHCP_DISCOVER =  1,                 /* DHCP discover packet */
    DHCP_OFFER = 2,                     /* DHCP offer packet */
    DHCP_REQUEST = 3,                   /* DHCP request packet */
    DHCP_DECLINE = 4,                   /* DHCP invalid config */
    DHCP_ACK = 5,                       /* DHCP acknowledge packet */
    DHCP_NAK = 6,                       /* DHCP negative ack packet */
    DHCP_RELEASE = 7,                   /* DHCP release IP addr */
    DHCP_INFORM =  8                    /* DHCP request information */  
} dhcp_msg_type_t;

/* Timeout control */
enum {
    INIT_TIMEOUT = 0,
    NEXT_TIMEOUT = 1,
};

/* Option index in the DHCP message option field */
typedef enum dhcp_opt_index {
    DHC_OPT_MSG_TYPE =  0,              /* DHCP message type */
    DHC_OPT_SERVER_INADDR = 1,          /* DHCP server IP address */
    DHC_OPT_LEASE_TIME = 2,             /* lease time */
    DHC_OPT_RENEWAL_TIME = 3,           /* renewal time */
    DHC_OPT_REBIND_TIME = 4,            /* rebind time */
    DHC_OPT_NETMASK = 5,                /* netmask */
    DHC_OPT_BCAST_INADDR = 6,           /* broadcast address */
    DHC_OPT_DHCP_MESSAGE = 7,           /* DHCP message */
    DHC_OPT_DHCP_CLASS_ID = 8,          /* DHCP class identifier */
    DHC_OPT_NTP_SERVER = 9,             /* NTP server's IP address */
    DHC_OPT_TIME_SERVER = 10,           /* time server's IP address */
    DHC_OPT_DNS = 11,                   /* Domain Name Server's IP address */
    DHC_OPT_LPR_SERVER = 12,            /* lpr server's IP address */
    DHC_OPT_HOST_NAME = 13,             /* hostname */
    DHC_OPT_DOMAIN_NAME = 14,           /* domainname */
    DHC_OPT_NISDOM_NAME = 15,           /* NIS domainname */
    DHC_OPT_ROUTER = 16,                /* routers on the client's subnet */
    N_SUPPORT_OPTIONS = 17              /* number of supported options */
} dhcp_opt_index_t;

/* Timeout control */
typedef enum {
    DISABLED = 0,
    ENABLED = 1,
} switch_t;

/* Define some data structure we can map onto the options field, making
   it easier to access the data values for each option. These options
   are defined in RFC1533, RFC2131, RFC2133. */

/* This is a template for all the Server Name options. These are
 all an option byte, followed by a length byte, followed by an 
 array of 2 IP addresses. */
typedef struct dhcp_opt_template {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip[DHCP_ADDR_LIST_SIZE];      /* array of IP addresses */
} dhcp_opt_template_t;

/* This is a template for all the options that require a filename */
typedef struct dhcp_opt_name_template {
    u_int8_t opt;
    u_int8_t len;
    char *filename[DHCP_FILESPEC_SIZE];
} dhcp_opt_name_template_t;

/* section 3.3 Subnet Mask
 *
 * The subnet mask option specifies the client's subnet mask as per
 * RFC 950 [5]. If both the subnet mask and the router option are
 * specified in a DHCP reply, the subnet mask option MUST be first. 
 *
 * The code for the subnet mask option is 1, and its length is 4 octets. */
typedef struct dhcp_subnet_mask {
    u_int8_t opt;
    u_int8_t len;
    ip4_t mask;
} dhcp_opt_subnet_mask_t;

/* section 3.4 Time Offset
 *
 * The time offset field specifies the offset of the client's subnet in
 * seconds from Coordinated Universal Time (UTC). The offset is
 * expressed as a two's complement 32-bit integer. A positive offset
 * indicates a location east of the zero meridian and a negative offset
 * indicates a location west of the zero meridian. 
 *
 * The code for the time offset option is 2, and its length is 4 octets. */
typedef struct dhcp_time_off {
    u_int8_t opt;
    u_int8_t len;
    u_int32_t time;
} dhcp_time_off_t;

/* section 3.5 Router Option
 *
 * The router option specifies a list of IP addresses for routers on the
 * client's subnet. Routers SHOULD be listed in order of preference. 
 *
 * The code for the router option is 3. The minimum length for the router
 * option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_opt_router_t;

/* section 3.6 Time Server Option
 *
 * The time server option specifies a list of RFC 868 [6] time servers
 * available to the client. Servers SHOULD be listed in order of
 * preference. 
 *
 * The code for the time server option is 4. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_opt_time_server_t;

/* section 3.7 Name Server Option
 *
 * The name server option specifies a list of IEN 116 [7] name servers
 * available to the client. Servers SHOULD be listed in order of
 * preference. 
 *
 * The code for the name server option is 5. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_opt_name_server_t;

/* section 3.8. Domain Name Server Option
 *
 * The domain name server option specifies a list of Domain Name System
 * (STD 13, RFC 1035 [8]) name servers available to the client. Servers
 * SHOULD be listed in order of preference.
 *
 * The code for the domain name server option is 6. The minimum length for
 * this option is 4 octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_opt_dns_t;

/* section 3.9. Log Server Option
 *
 * The log server option specifies a list of MIT-LCS UDP log servers
 * available to the client. Servers SHOULD be listed in order of
 *  preference. 
 *
 * The code for the log server option is 7. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_opt_log_server_t;

/* section 3.10. Cookie Server Option
 *
 * The cookie server option specifies a list of RFC 865 [9] cookie servers
 * available to the client. Servers SHOULD be listed in order of
 * preference. 
 *
 * The code for the log server option is 8. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_opt_cookie_server_t;

/* section 3.11. LPR Server Option
 *
 * The LPR server option specifies a list of RFC 1179 [10] line printer
 * servers available to the client. Servers SHOULD be listed in order of
 * preference. 
 *
 * The code for the LPR server option is 9. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_opt_lpr_server_t;

/* section 3.12. Impress Server Option
 *
 * The Impress server option specifies a list of Imagen Impress servers
 * available to the client. Servers SHOULD be listed in order of
 * preference. 
 *
 * The code for the Impress server option is 10. The minimum length for
 * this option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_opt_impress_server_t;

/* section 3.13. Resource Location Server Option
 *
 * This option specifies a list of RFC 887 [11] Resource Location servers
 * available to the client. Servers SHOULD be listed in order of
 * preference. 
 *
 * The code for this option is 11. The minimum length for this option is 4
 * octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_opt_res_loc_server_t;

/* section 3.14. Host Name Option
 *
 * This option specifies the name of the client. The name may or may not be
 * qualified with the local domain name (see section 3.17 for the preferred
 * way to retrieve the domain name). See RFC 1035 for character set
 * restrictions.
 *
 * The code for this option is 12, and its minimum length is 1. 
 */
typedef struct dhcp_opt_name_template dhcp_opt_hostname_t;

/* section 3.15. Boot File Size Option
 *
 * This option specifies the length in 512-octet blocks of the default boot
 * image for the client. The file length is specified as an unsigned 16-bit
 * integer.
 *
 * The code for this option is 13, and its length is 2.
 */
typedef struct dhcp_opt_bootfile_size {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 2 */
    u_int16_t size;
} dhcp_opt_bootfile_size_t;

/* section 3.16. Merit Dump File
 *
 * This option specifies the path-name of a file to which the client's core
 * image should be dumped in the event the client crashes. The path is
 * formatted as a character string consisting of characters from the NVT
 * ASCII character set.  
 *
 * The code for this option is 14. Its minimum length is 1.
 */
typedef struct dhcp_opt_name_template dhcp_opt_dumpfile_t;

/* section 3.17. Domain Name
 *
 * This option specifies the domain name that client should use when
 * resolving hostnames via the Domain Name System. 
 *
 * The code for this option is 15. Its minimum length is 1. 
 */
typedef struct dhcp_opt_name_template dhcp_opt_domainname_t;

/* section 3.18. Swap Server
 *
 * This specifies the IP address of the client's swap server.
 *
 * The code for this option is 16 and its length is 4. 
 */
typedef struct dhcp_opt_swap_server {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip;
} dhcp_opt_swap_server_t;

/* section 3.19. Root Path
 *
 * This option specifies the path-name that contains the client's root
 * disk. The path is formatted as a character string consisting of
 * characters from the NVT ASCII character set.  
 *
 * The code for this option is 17. Its minimum length is 1.
 */
typedef struct dhcp_opt_name_template dhcp_opt_root_path_t;

/* section 3.20. Extensions Path
 *
 * A string to specify a file, retrievable via TFTP, which contains
 * information which can be  interpreted in the same way as the 64-octet
 * vendor-extension field within the BOOTP response, with the following
 * exceptions: 
 *
 *    - the length of the file is unconstrained;
 *    - all references to Tag 18 (i.e., instances of the
 *      BOOTP Extensions Path field) within the file are ignored.
 *
 * The code for this option is 18. Its minimum length is 1. 
 */
typedef struct dhcp_opt_name_template dhcp_opt_ext_path_t;

/* section 4.1. IP Forwarding Enable/Disable Option
 *
 * This option specifies whether the client should configure its IP layer
 * for packet forwarding. A value of 0 means disable IP forwarding, and a
 * value of 1 means enable IP forwarding. 
 *
 * The code for this option is 19, and its length is 1.
 */
typedef struct dhcp_opt_ip_forward {
    u_int8_t opt;
    u_int8_t len;
    switch_t flag;
} dhcp_opt_ip_forward_t;

/* section 4.2. Non-Local Source Routing Enable/Disable Option
 *
 * This option specifies whether the client should configure its IP layer
 * to allow forwarding of datagrams with non-local source routes. A value
 * of 0 means disallow forwarding of such datagrams, and a value of 1
 * means allow forwarding. 
 *
 * The code for this option is 20, and its length is 1.
 */
typedef struct dhcp_opt_nonlocal_routing {
    u_int8_t opt;
    u_int8_t len;
    switch_t flag;
} dhcp_opt_nonlocal_routing_t;

/* section 4.3. Policy Filter Option
 *
 * This option specifies policy filters for non-local source routing. The
 * filters consist of a list of IP addresses and masks which specify
 * destination/mask pairs with which to filter incoming source routes. 
 * Any source routed datagram whose next-hop address does not match one of
 * the filters should be discarded by the client.
 *
 * The code for this option is 21. The minimum length of this option is 8,
 * and the length MUST be a multiple of 8. 
 */
typedef struct dhcp_opt_policy_filter {
    u_int8_t opt;
    u_int8_t len;
    ip4_t addr1;
    u_int32_t mask1;
    ip4_t addr2;
    u_int32_t mask2;
} dhcp_opt_policy_filter_t;

/* section 4.4. Maximum Datagram Reassembly Size
 *
 * This option specifies the maximum size datagram that the client should be
 * prepared to reassemble. The size is specified as a 16-bit unsigned
 * integer. The minimum value legal value is 576. 
 *
 * The code for this option is 22, and its length is 2.
 */
typedef struct dhcp_opt_datagram_size {
    u_int8_t opt;
    u_int8_t len;                       /* will always be 2 */
    u_int16_t size;
} dhcp_opt_datagram_size_t;

/* section 4.5. Default IP Time-to-live
 *
 * This option specifies the default time-to-live that the client should
 * use on outgoing datagrams. The TTL is specified as an octet with a
 * value between 1 and 255. 
 *
 * The code for this option is 23, and its length is 1.
 */
typedef struct dhcp_opt_time_to_live {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 1 */
    u_int8_t time;
} dhcp_opt_time_to_live_t;

/* section 4.6. Path MTU Aging Timeout Option
 *
 * This option specifies the timeout (in seconds) to use when aging Path
 * MTU values discovered by the mechanism defined in RFC 1191 [12]. The
 * timeout is specified as a 32-bit unsigned integer. 
 *
 * The code for this option is 24, and its length is 4. 
 */
typedef struct dhcp_opt_mtu_timeout {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 4 */
    u_int32_t timeout;
} dhcp_opt_mtu_timeout_t;

/* section 4.7. Path MTU Plateau Table Option
 *
 * This option specifies a table of MTU sizes to use when performing Path
 * MTU Discovery as defined in RFC 1191. The table is formatted as a list
 * of 16-bit unsigned integers, ordered from smallest to largest. The
 * minimum MTU value cannot be smaller than 68.
 *
 * The code for this option is 25. Its minimum length is 2, and the length
 * MUST be a multiple of 2.
 */
typedef struct dhcp_opt_mtu_plateau {
    u_int8_t opt;
    u_int8_t len;
    u_int16_t plateau[DHCP_MIN_OPT_SIZE];
} dhcp_opt_mtu_plateau_t;

/* section 5.1. Interface MTU Option
 *
 * This option specifies the MTU to use on this interface. The MTU is
 * specified as a 16-bit unsigned integer. The minimum legal value for the
 * MTU is 68.  
 *
 * The code for this option is 26, and its length is 2.
 */
typedef struct dhcp_opt_mtu {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 2 */
    u_int16_t mtu;
} dhcp_opt_mtu_t;

/* section 5.2. All Subnets are Local Option
 *
 * This option specifies whether or not the client may assume that all
 * subnets of the IP network to which the client is connected use the same
 * MTU as the subnet of that network to which the client is directly
 * connected. A value of 1 indicates that all subnets share the same MTU.
 * A value of 0 means that the client should assume that some subnets of
 * the directly connected network may have smaller MTUs. 
 *
 * The code for this option is 27, and its length is 1. 
 */
typedef struct dhcp_opt_local_subnets {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 1 */
    switch_t flag;
} dhcp_opt_local_subnets_t;

/* section 5.3. Broadcast Address Option
 *
 * This option specifies the broadcast address in use on the client's
 * subnet. Legal values for broadcast addresses are specified in section
 * 3.2.1.3 of [4]. 
 *
 * The code for this option is 28, and its length is 4.
 */
typedef struct dhcp_opt_broadcast_addr {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 4 */
    ip4_t addr;
} dhcp_opt_broadcast_addr_t;

/* section 5.4. Perform Mask Discovery Option
 *
 * This option specifies whether or not the client should perform subnet mask
 * discovery using ICMP. A value of 0 indicates that the client should not
 * perform mask discovery. A value of 1 means that the client should perform
 * mask discovery.
 *
 * The code for this option is 29, and its length is 1.
 */
typedef struct dhcp_opt_mask_discovery {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 1 */
    switch_t flag;
} dhcp_opt_mask_discovery_t;

/* section 5.5. Mask Supplier Option
 *
 * This option specifies whether or not the client should respond to subnet
 * mask requests using ICMP. A value of 0 indicates that the client should
 * not respond. A value of 1 means that the client should respond. 
 *
 * The code for this option is 30, and its length is 1.
 */
typedef struct dhcp_opt_mask_supplier {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 1 */
    switch_t flag;
} dhcp_opt_mask_supplier_t;

/* section 5.6. Perform Router Discovery Option
 *
 * This option specifies whether or not the client should solicit routers
 * using the Router Discovery mechanism defined in RFC 1256 [13]. A value
 * of 0 indicates that the client should not perform router discovery. A
 * value of 1 means that the client should perform router discovery. 
 *
 * The code for this option is 31, and its length is 1.
 */
typedef struct dhcp_opt_router_discovery {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 1 */
    switch_t flag;
} dhcp_opt_router_discovery_t;

/* section 5.7. Router Solicitation Address Option
 *
 * This option specifies the address to which the client should transmit
 * router solicitation requests.
 *
 * The code for this option is 32, and its length is 4.
 */
typedef struct dhcp_opt_router_addr {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 4 */
    u_int32_t addr;
} dhcp_opt_router_addr_t;

/* section 5.8. Static Route Option
 *
 * This option specifies a list of static routes that the client should
 * install in its routing cache. If multiple routes to the same destination
 * are specified, they are listed in descending order of priority. 
 *
 * The routes consist of a list of IP address pairs. The first address is
 * the destination address, and the second address is the router for the
 * destination. 
 * The default route (0.0.0.0) is an illegal destination for a static
 * route. See section 3.5 for information about the router option. 
 *
 * The code for this option is 33. The minimum length of this option is 8,
 * and the length MUST be a multiple of 8.
 */
typedef struct dhcp_opt_static_route {
    u_int8_t opt;
    u_int8_t len;                       /* this must be a multiple of 8 */
    union {
        u_int32_t destination;
        u_int32_t router;
    } u[10];
} dhcp_opt_static_route_t;

/* section 6.1. Trailer Encapsulation Option 
 *
 * This option specifies whether or not the client should negotiate
 * the use of trailers (RFC 893 [14]) when using the ARP protocol.
 * A value of 0 indicates that the client should not attempt to use  
 * trailers. A value of 1 means that the client should attempt to use
 * trailers.  
 *
 * The code for this option is 34, and its length is 1.
 */
typedef struct dhcp_opt_trailer_encaps {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 1 */
    switch_t flag;
} dhcp_opt_trailer_encaps_t;

/* section 6.2. ARP Cache Timeout Option
 *
 * This option specifies the timeout in seconds for ARP cache entries. The
 * time is specified as a 32-bit unsigned integer. 
 *
 * The code for this option is 35, and its length is 4.
 */
typedef struct dhcp_opt_arp_timeout {
    u_int8_t opt;
    u_int8_t len;
    u_int32_t timeout;
} dhcp_opt_arp_timeout_t;

/* section 6.3. Ethernet Encapsulation Option
 *
 * This option specifies whether or not the client should use Ethernet
 * Version 2 (RFC 894 [15]) or IEEE 802.3 (RFC 1042 [16]) encapsulation
 * if the interface is an Ethernet. A value of 0 indicates that the client
 * should use RFC 894 encapsulation. A value of 1 means that the client
 * should use RFC 1042 encapsulation. 
 *
 * The code for this option is 36, and its length is 1. 
 */
typedef struct dhcp_opt_eth_encaps {
    u_int8_t opt;
    u_int8_t len;                       /* this will always be 1 */
    switch_t flag;
} dhcp_opt_eth_encaps_t;

/* section 7.1. TCP Default TTL Option
 *
 * This option specifies the default TTL that the client should use when
 * sending TCP segments. The value is represented as an 8-bit unsigned
 * integer. The minimum value is 1.  
 *
 * The code for this option is 37, and its length is 1.
 */
typedef struct dhcp_opt_tcp_default_ttl {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t ttl;
} dhcp_opt_tcp_default_ttl_t;

/* section 7.2. TCP Keepalive Interval Option
 *
 * This option specifies the interval (in seconds) that the client TCP
 * should wait before sending a keepalive message on a TCP connection.
 * The time is specified as a 32-bit unsigned integer. A value of zero
 * indicates that the client should not generate keepalive messages on
 * connections unless specifically requested by an application. 
 *
 * The code for this option is 38, and its length is 4.
 */
typedef struct dhcp_opt_tcp_keepalive_int {
    u_int8_t opt;
    u_int8_t len;
    u_int32_t time;
} dhcp_opt_tcp_keepalive_int_t;

/* section 7.3. TCP Keepalive Garbage Option
 *
 * This option specifies the whether or not the client should send TCP
 * keepalive messages with a octet of garbage for compatibility with
 * older implementations. A value of 0 indicates that a garbage octet
 * should not be sent. A value of 1 indicates that a garbage octet should
 * be sent.
 *
 * The code for this option is 39, and its length is 1. 
 */
typedef struct dhcp_opt_tcp_keepalive_garbage {
    u_int8_t opt;
    u_int8_t len;
    switch_t flag;
} dhcp_opt_tcp_keepalive_garbage_t;

/* section 8.1. Network Information Service Domain Option
 *
 * This option specifies the name of the client's NIS [17] domain. The
 * domain is formatted as a character string consisting of characters
 * from the NVT ASCII character set. 
 *
 * The code for this option is 40. Its minimum length is 1. 
 */
typedef struct dhcp_opt_name_template dhcp_opt_nis_domain_t;

/* section 8.2. Network Information Servers Option
 *
 * This option specifies a list of IP addresses indicating NIS servers
 * available to the client. Servers SHOULD be listed in order of
 * preference. 
 *
 * The code for this option is 41. Its minimum length is 4, and the length
 * MUST be a multiple of 4.
 */
typedef struct dhcp_opt_nis_servers {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip[DHCP_ADDR_LIST_SIZE];
} dhcp_opt_nis_servers_t;

/* section 8.3. Network Time Protocol Servers Option
 *
 * This option specifies a list of IP addresses indicating NTP [18]
 * servers available to the client. Servers SHOULD be listed in order of
 * preference.  
 *
 * The code for this option is 42. Its minimum length is 4, and the length
 * MUST be a multiple of 4. 
 */
typedef struct dhcp_opt_ntp_servers {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip[DHCP_ADDR_LIST_SIZE];
} dhcp_opt_ntp_servers_t;

/* section 8.4. Vendor Specific Information
 *
 * This option is used by clients and servers to exchange vendor- specific
 * information. The information is an opaque object of n octets, presumably
 * interpreted by vendor-specific code on the clients and servers. The
 * definition of this information is vendor specific. The vendor is indicated
 * in the vendor class identifier option. Servers not equipped to interpret
 * the vendor-specific information sent by a client MUST ignore it (although
 * it may be reported).  Clients which do not receive desired vendor-specific
 * information SHOULD make an attempt to operate without it, although they may
 * do so (and announce they are doing so) in a degraded mode.
 * If a vendor potentially encodes more than one item of information in
 * this option, then the vendor SHOULD encode the option using "Encapsulated
 * vendor-specific options" as described below: The Encapsulated
 * vendor-specific options field SHOULD be encoded as a sequence of
 * code/length/value fields of identical syntax to the DHCP options field with
 * the following exceptions:
 *
 *    1) There SHOULD NOT be a "magic cookie" field in the encapsulated
 *       vendor-specific extensions field. 
 *    2) Codes other than 0 or 255 MAY be redefined by the vendor within the
 *       encapsulated vendor-specific extensions field, but SHOULD conform to
 *       the tag-length-value syntax defined in section 2. 
 *    3) Code 255 (END), if present, signifies the end of the encapsulated
 *       vendor extensions, not the end of the vendor extensions field. If no
 *       code 255 is present, then the end of the enclosing vendor-specific
 *       information field is taken as the end of the encapsulated
 *       vendor-specific extensions field. 
 *
 * The code for this option is 43 and its minimum length is 1.
 */
typedef struct dhcp_opt_vendor_specific {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t vendor_opts[DHCP_MAX_OPT_SIZE];
} dhcp_opt_vendor_specific_t;

/* section 8.5. NetBIOS over TCP/IP Name Server Option
 *
 * The NetBIOS name server (NBNS) option specifies a list of RFC 1001/1002
 * [19] [20] NBNS name servers listed in order of preference. 
 *
 * The code for this option is 44. The minimum length of the option is 4
 * octets, and the length must always be a multiple of 4. 
 */
typedef struct dhcp_opt_nbns {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip[DHCP_ADDR_LIST_SIZE];
} dhcp_opt_nbns_t;

/* section 8.6. NetBIOS over TCP/IP Datagram Distribution Server Option
 *
 * The NetBIOS datagram distribution server (NBDD) option specifies a list of
 * RFC 1001/1002 NBDD servers listed in order of preference.
 *
 * The code for this option is 45. The minimum length of the option is 4
 * octets, and the length must always be a multiple of 4. 
 */
typedef struct dhcp_opt_nbdd {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip[DHCP_ADDR_LIST_SIZE];
} dhcp_opt_nbdd_t;

/* section 8.7. NetBIOS over TCP/IP Node Type Option
 *
 * The NetBIOS node type option allows NetBIOS over TCP/IP clients which are
 * configurable to be configured as described in RFC 1001/1002. The value is
 * specified as a single octet which identifies the client type as follows: 
 *
 *            Value         Node Type
 *            -----         ---------
 *            0x1           B-node
 *            0x2           P-node
 *            0x4           M-node
 *            0x8           H-node
 * In the above chart, the notation '0x' indicates a number in base-16
 * (hexadecimal).  
 *
 * The code for this option is 46. The length of this option is always 1. 
 */
typedef struct dhcp_opt_netbios_node_type {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t type;
} dhcp_opt_netbios_node_type_t;

/* section 8.8. NetBIOS over TCP/IP Scope Option
 *
 * The NetBIOS scope option specifies the NetBIOS over TCP/IP scope parameter
 * for the client as specified in RFC 1001/1002. See [19], [20], and [8] for
 * character-set restrictions.  
 *
 * The code for this option is 47. The minimum length of this option is 1.
 */
typedef struct dhcp_opt_netbios_scope {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t scope[DHCP_MAX_OPT_SIZE];
} dhcp_opt_netbios_scope_t;

/* section 8.9. X Window System Font Server Option
 *
 * This option specifies a list of X Window System [21] Font servers available
 * to the client. Servers SHOULD be listed in order of preference. 
 *
 * The code for this option is 48. The minimum length of this option is 4
 * octets, and the length MUST be a multiple of 4.
 */
typedef struct dhcp_opt_x_font {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip[DHCP_ADDR_LIST_SIZE];
} dhcp_opt_x_font_t;

/* section 8.10. X Window System Display Manager Option
 *
 * This option specifies a list of IP addresses of systems that are running
 * the X Window System Display Manager and are available to the client.
 * Addresses SHOULD be listed in order of preference. 
 *
 * The code for the this option is 49. The minimum length of this option is 4,
 * and the length MUST be a multiple of 4. 
 */
typedef struct dhcp_opt_x_display_mgr {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip[DHCP_ADDR_LIST_SIZE];
} dhcp_opt_x_display_mgr_t;


/* 8.11. Network Information Service+ Domain Option
 *
 * This option specifies the name of the client's NIS+ [17] domain. The domain
 * is formatted as a character string consisting of characters from the NVT
 * ASCII character set.  
 *
 * The code for this option is 64. Its minimum length is 1. 
 */
typedef struct dhcp_opt_template dhcp_nis_domain_t;

/* 8.12. Network Information Service+ Servers Option
 *
 * This option specifies a list of IP addresses indicating NIS+ servers
 * available to the client. Servers SHOULD be listed in order of preference. 
 *
 * The code for this option is 65. Its minimum length is 4, and the length
 * MUST be a multiple of 4.  
 */
typedef struct dhcp_opt_template dhcp_nis_service_t;


/* 8.13. Mobile IP Home Agent option
 *
 * This option specifies a list of IP addresses indicating mobile IP home
 * agents available to the client. Agents SHOULD be listed in order of
 * preference.  
 *
 * The code for this option is 68. Its minimum length is 0 (indicating no home
 * agents are available) and the length MUST be a multiple of 4. It is
 * expected that the usual length will be four octets, containing a single
 * home agent's address.  
 */
typedef struct dhcp_opt_template dhcp_mbile_ip_t;

/* 8.14. Simple Mail Transport Protocol (SMTP) Server Option
 *
 * The SMTP server option specifies a list of SMTP servers available to the
 * client. Servers SHOULD be listed in order of preference. 
 *
 * The code for the SMTP server option is 69. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_smtp_server_t;

/* 8.15. Post Office Protocol (POP3) Server Option
 *
 * The POP3 server option specifies a list of POP3 available to the
 * client. Servers SHOULD be listed in order of preference. 
 *
 * The code for the POP3 server option is 70. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_pop3_server_t;

/* 8.16. Network News Transport Protocol (NNTP) Server Option
 *
 * The NNTP server option specifies a list of NNTP available to the
 * client. Servers SHOULD be listed in order of preference. 
 *
 * The code for the NNTP server option is 71. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_nntp_server_t;

/* 8.17. Default World Wide Web (WWW) Server Option
 *
 * The WWW server option specifies a list of WWW available to the
 * client. Servers SHOULD be listed in order of preference. 
 * 
 * The code for the WWW server option is 72. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_http_server_t;

/* 8.18. Default Finger Server Option
 *
 * The Finger server option specifies a list of Finger available to the
 * client. Servers SHOULD be listed in order of preference. 
 *
 * The code for the Finger server option is 73. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_finger_server_t;

/* 8.19. Default Internet Relay Chat (IRC) Server Option
 *
 * The IRC server option specifies a list of IRC available to the
 * client. Servers SHOULD be listed in order of preference. 
 *
 * The code for the IRC server option is 74. The minimum length for this
 * option is 4 octets, and the length MUST always be a multiple of 4. 
 */
typedef struct dhcp_opt_template dhcp_irc_server_t;

/* 8.20. StreetTalk Server Option
 *
 * The StreetTalk server option specifies a list of StreetTalk servers
 * available to the client. Servers SHOULD be listed in order of preference. 
 *
 * The code for the StreetTalk server option is 75. The minimum length for
 * this option is 4 octets, and the length MUST always be a multiple of 4.
 */
typedef struct dhcp_opt_template dhcp_street_server_t;

/* 8.21. StreetTalk Directory Assistance (STDA) Server Option
 *
 * The StreetTalk Directory Assistance (STDA) server option specifies a list
 * of STDA servers available to the client. Servers SHOULD be listed in order
 * of preference.  
 *
 * The code for the StreetTalk Directory Assistance server option is 76. The
 * minimum length for this option is 4 octets, and the length MUST always be a
 * multiple of 4.  
 */
typedef struct dhcp_opt_template dhcp_stda_ass_t;

/* All these following options are DHCP specific extentions to BOOTP */

/* section 9.1. Requested IP Address
 *
 * This option is used in a client request (DHCPDISCOVER) to allow the client
 * to request that a particular IP address be assigned. 
 *
 * The code for this option is 50, and its length is 4.
 */
typedef struct dhcp_opt_request_ip {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip;
} dhcp_opt_request_ip_t;

/* section 9.2. IP Address Lease Time
 *
 * This option is used in a client request (DHCPDISCOVER or DHCPREQUEST) to
 * allow the client to request a lease time for the IP address. In a server
 * reply (DHCPOFFER), a DHCP server uses this option to specify the lease time
 * it is willing to offer.  
 * The time is in units of seconds, and is specified as a 32-bit unsigned
 * integer.  
 *
 * The code for this option is 51, and its length is 4.
 */
typedef struct dhcp_opt_lease_time {
    u_int8_t opt;
    u_int8_t len;
    u_int32_t time;
} dhcp_opt_lease_time_t;

/* section 9.3. Option Overload
 *
 * This option is used to indicate that the DHCP 'sname' or 'file' fields are
 * being overloaded by using them to carry DHCP options. A DHCP server inserts
 * this option if the returned parameters will exceed the usual space allotted
 * for options.
 * If this option is present, the client interprets the specified
 * additional fields after it concludes interpretation of the standard option
 * fields.
 *
 * The code for this option is 52, and its length is 1. Legal values for this
 * option are:  
 *
 *    Value   Meaning
 *    -----   --------
 *      1     the 'file' field is used to hold options
 *      2     the 'sname' field is used to hold options
 *      3     both fields are used to hold options
 */
typedef struct dhcp_opt_overload {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t flag;
} dhcp_opt_overload_t;

/* 9.4 TFTP server name
 *
 * This option is used to identify a TFTP server when the 'sname' field in the
 * DHCP header has been used for DHCP options. 
 *
 * The code for this option is 66, and its minimum length is 1. 
 */
typedef struct dhcp_opt_name_template dhcp_tftp_server_t;

/* 9.5 Bootfile name
 *
 * This option is used to identify a bootfile when the 'file' field in the
 *  DHCP header has been used for DHCP options. 
 *
 * The code for this option is 67, and its minimum length is 1.
 */
typedef struct dhcp_opt_name_template dhcp_bootfile_t;

/* 9.6 DHCP Message Type
 *
 * This option is used to convey the type of the DHCP message. The code for
 * this option is 53, and its length is 1. Legal values for this option are:
 * 
 *     Value   Message Type
 *     -----   ------------
 *       1     DHCPDISCOVER
 *       2     DHCPOFFER
 *       3     DHCPREQUEST
 *       4     DHCPDECLINE
 *       5     DHCPACK
 *       6     DHCPNAK
 *       7     DHCPRELEASE
 *       8     DHCPINFORM
 */
typedef struct dhcp_opt_msg_type {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t type;
//    dhcp_msg_type_t type;
} dhcp_opt_msg_type_t;

/* 9.7. Server Identifier
 *
 * This option is used in DHCPOFFER and DHCPREQUEST messages, and may
 * optionally be included in the DHCPACK and DHCPNAK messages. DHCP servers
 * include this option in the DHCPOFFER in order to allow the client to
 * distinguish between lease offers. DHCP clients use the contents of the
 * 'server identifier' field as the destination address for any DHCP messages
 * unicast to the DHCP server. DHCP clients also indicate which of several
 * lease offers is being accepted by including this option in a DHCPREQUEST
 * message.  The identifier is the IP address of the selected server.
 *
 * The code for this option is 54, and its length is 4.
 */
typedef struct dhcp_opt_server_id {
    u_int8_t opt;
    u_int8_t len;
    ip4_t ip;
} dhcp_opt_server_id_t;

/* 9.8. Parameter Request List
 *
 * This option is used by a DHCP client to request values for specified
 * configuration parameters.  The list of requested parameters is specified as
 * n octets, where each octet is a valid DHCP option code as defined in this
 * document.  The client MAY list the options in order of preference. The DHCP
 * server is not required to return the options in the requested order, but
 * MUST try to insert the requested options in the order requested by the
 * client.
 *
 * The code for this option is 55. Its minimum length is 1.
 */
typedef struct dhcp_opt_param_list {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t codes[DHCP_MAX_OPT_SIZE];
} dhcp_opt_param_list_t;

/* 9.9. Message
 *
 * This option is used by a DHCP server to provide an error message to a DHCP
 * client in a DHCPNAK message in the event of a failure. A client may use
 * this option in a DHCPDECLINE message to indicate the why the client
 * declined the offered parameters. The message consists of n octets of NVT
 * ASCII text, which the client may display on an available output device.
 *
 * The code for this option is 56 and its minimum length is 1.
 */
typedef struct dhcp_opt_message {
    u_int8_t opt;
    u_int8_t len;
    char text[DHCP_MAX_OPT_SIZE];
} dhcp_opt_message_t;

/* 9.10. Maximum DHCP Message Size
 *
 * This option specifies the maximum length DHCP message that it is willing to
 * accept. The length is specified as an unsigned 16-bit integer. A client may
 * use the maximum DHCP message size option in DHCPDISCOVER or DHCPREQUEST
 * messages, but should not use the option in DHCPDECLINE messages. 
 *
 * The code for this option is 57, and its length is 2. The minimum legal
 * value is 576 octets.
 */
typedef struct dhcp_opt_max_msg_size {
    u_int8_t opt;
    u_int8_t len;
    u_int16_t size;
} dhcp_opt_max_msg_size_t;

/* 9.11. Renewal (T1) Time Value
 *
 * This option specifies the time interval from address assignment until the
 * client transitions to the RENEWING state. The value is in units of seconds,
 * and is specified as a 32-bit unsigned integer.  
 *
 * The code for this option is 58, and its length is 4.
 */
typedef struct dhcp_opt_renewal_time {
    u_int8_t opt;
    u_int8_t len;
    u_int32_t time;
} dhcp_opt_renewal_time_t;

/* 9.12. Rebinding (T2) Time Value
 *
 * This option specifies the time interval from address assignment until the
 * client transitions to the REBINDING state. The value is in units of
 * seconds, and is specified as a 32-bit unsigned integer.
 *
 * The code for this option is 59, and its length is 4.
 */
typedef struct dhcp_opt_rebinding_time {
    u_int8_t opt;
    u_int8_t len;
    u_int32_t time;
} dhcp_opt_rebinding_time_t;

/* 9.13. Class-identifier
 *
 * This option is used by DHCP clients to optionally identify the vendor type
 * and configuration of a DHCP client. The information is a string of n
 * octets, interpreted by servers. Vendors may choose to define specific
 * vendor class identifiers to convey particular configuration or other
 * identification information about a client. For example, the identifier may
 * encode the client's hardware configuration. Servers not equipped to
 * interpret the class-specific information sent by a client MUST ignore it
 * (although it may be reported). Servers that respond SHOULD only use option
 * 43 to return the vendor-specific information to the client.
 *
 * The code for this option is 60, and its minimum length is 1.
 */
typedef struct dhcp_opt_class_id {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t id[1];
} dhcp_opt_class_id_t;

/* 9.14. Client-identifier
 *
 * This option is used by DHCP clients to specify their unique
 * identifier. DHCP servers use this value to index their database of address
 * bindings. This value is expected to be unique for all clients in an
 * administrative domain.  Identifiers SHOULD be treated as opaque objects by
 * DHCP servers.  The client identifier MAY consist of type-value pairs
 * similar to the 'htype'/'chaddr' fields defined in [3]. For instance, it MAY
 * consist of a hardware type and hardware address. In this case the type
 * field SHOULD be one of the ARP hardware types defined in STD2 [22]. A
 * hardware type of 0 (zero) should be used when the value field contains an
 * identifier other than a hardware address (e.g. a fully qualified domain
 * name).  For correct identification of clients, each client's client-
 * identifier MUST be unique among the client-identifiers used on the subnet
 * to which the client is attached. Vendors and system administrators are
 * responsible for choosing client-identifiers that meet this requirement for
 * uniqueness.
 *
 * The code for this option is 61, and its minimum length is 2.
 */
typedef struct dhcp_opt_client_id {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t id[1];
} dhcp_opt_client_id_t;

typedef enum dhcp_sysarch_types_s {
    DHCP_STD_X86 = 0,
    DHCP_PC98_X86 = 1,
    DHCP_IA64PC = 2,
    DHCP_DEC_ALPHA = 3,
    DHCP_ARC_X86 = 4,
    DHCP_LEAN = 5
} dhcp_sysarch_types_t;

typedef struct dhcp_sysarch_s {
    u_int8_t opt;
    u_int8_t len;
    u_int16_t type;
} dhcp_opt_sysarch_t;

#define DHCP_UNDI 0x01                  /* 2 bytes of data */
#define DHCP_PCI 0x02                   /* 12 bytes of data */
#define DHCP_PNP 0x03                   /* 7 bytes of data */
typedef struct dhcp_nicif {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t type;
    u_int8_t major;
    u_int8_t minor;
    u_int8_t tag;
} dhcp_opt_nicif_t;

#define DHCP_PLATFORMID_GUID 0
typedef struct dhcp_platformid {
    u_int8_t opt;
    u_int8_t len;
    u_int8_t type;
    u_int8_t buf[1];
} dhcp_opt_platformid_t;

#if 0
int32_t
fill_dhcp_request(dhcp_msg_type_t type, bootp_t *bootp_packet,
                  int8_t *macaddr, u_int32_t xid);
#endif

/* These options are commonly used in a DHCPDISCOVER message */
void add_dhcp_nicif_opt (dhcp_t *pkt);
void add_dhcp_msg_type_opt (dhcp_t *pkt);
void add_dhcp_max_msg_size_opt (dhcp_t *pkt);
void add_dhcp_platformid_opt (dhcp_t *pkt);
void add_dhcp_class_id_opt (dhcp_t *pkt);
void add_dhcp_client_id_opt (dhcp_t *pkt);
void add_dhcp_max_msg_len_opt (dhcp_t *pkt);
void add_dhcp_sysarch_opt (dhcp_t *pkt);
void add_dhcp_param_list_opt (dhcp_t *pkt);
void add_dhcp_nic_type_opt (dhcp_t *pkt);
void add_dhcp_request_ip_addr_opt (dhcp_t *pkt);
void add_dhcp_hostname_opt (dhcp_t *pkt);

/* These are the options commonly used in a DHCPOFFER mesage */
void add_dhcp_ipaddr_lease_time_opt (dhcp_t *pkt);
void add_dhcp_subnet_mask_opt (dhcp_t *pkt);
void add_dhcp_broadcast_addr_opt (dhcp_t *pkt);
void add_dhcp_routers_on_subnet_opt (dhcp_t *pkt);
void add_dhcp_domainname_opt (dhcp_t *pkt);
void add_dhcp_dns_opt (dhcp_t *pkt);
void add_dhcp_server_id_opt (dhcp_t *pkt);

#endif /* __DHCP_H__ */
