/* Nilo - A Network boot loader.

   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the tracing macros */
#include "trace.h"

/* Define the TFTP data structures */
#include "tftp.h"

/* Define the DHCP data structures */
#include "dhcp.h"

/* Define the ARP data structures */
#include "arp.h"

/* Define the  data structures */
#include "dump.h"

// #include <oskit/dev/ethernet.h>
// #include <oskit/dev/linux.h>
#include <net/if_arp.h>
#include <net/ether.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "pxenv.h"
#include "pxe.h"
#include "tagged.h"
#include "util.h"

/* Convert an internet number to it's ascified IP version */
char *
_inet_ntoa(unsigned int a) 
{
    struct in_addr i;
    
    //    NILO_TRACE_FUNCTION;
    
    i.s_addr = a;
    return (inet_ntoa(i));
    
    // NILO_TRACE_FUNCTION_RETURN;
}

void 
dump_pxenv (pxenv_t *pxenv)
{

    NILO_TRACE_MSG ("Dumping a PXENV+ structure.\r\n");

    if (pxenv == 0) {
        NILO_TRACE_MSG ("ERROR: PXENV+ structure is NULL\r\n");
        return;
    }
    if (memcmp (pxenv->signature, "PXENV+", 6)) {
        NILO_TRACE_MSG ("ERROR: PXENV+ structure is bogus\r\n");
        return;
    }

    NILO_TRACE_PRINTF ("\tpxenv.version: 0x%04x\r\n", pxenv->version);
    NILO_TRACE_PRINTF ("\tpxenv.length: 0x%02x\r\n", pxenv->struct_length);
    NILO_TRACE_PRINTF ("\tpxenv.cksum: 0x%02x\r\n", pxenv->struct_cksum);

    NILO_TRACE_PRINTF ("\tpxenv.rmentry: 0x%08x, pxenv->rmentry, pxenv.pmentry: 0x%08x\r\n", pxenv->rmentry, pxenv->pmentry);
    NILO_TRACE_PRINTF ("\tpxenv.stackseg: 0x%04x, pxenv.stacksize: 0x%04x\r\n", pxenv->stackseg, pxenv->stacksize);
    NILO_TRACE_PRINTF ("\tpxenv.basecode_seg: 0x%04x, pxenv.basecode_size: 0x%04x\r\n", pxenv->basecode_seg, pxenv->basecode_size);
    NILO_TRACE_PRINTF ("\tpxenv.basedata_seg: 0x%04x, pxenv.basedata_size: 0x%04x\r\n", pxenv->basedata_seg, pxenv->basedata_size);
    NILO_TRACE_PRINTF ("\tpxenv.undidata_seg: 0x%04x, pxenv.undidata_size: 0x%04x\r\n", pxenv->undidata_seg, pxenv->undidata_size);
    NILO_TRACE_PRINTF ("\tpxenv.undicode_seg: 0x%04x, pxenv.undicode_size: 0x%04x\r\n", pxenv->undicode_seg, pxenv->undicode_size);

    NILO_TRACE_PRINTF ("\tpxenv.pxeptr: 0x%08x\r\n", pxenv->pxeptr);
}

void dump_pxe (pxe_t *pxe)
{
    NILO_TRACE_MSG ("Dumping a !PXE structure.\r\n");

    if (pxe == 0) {
        NILO_TRACE_MSG ("ERROR: !PXE structure is NULL\r\n");
        return;
    }
    if (memcmp (pxe->signature, "!PXE", 4)) {
        NILO_TRACE_MSG ("ERROR: !PXE structure is bogus\r\n");
        return;
    }
    
    NILO_TRACE_PRINTF ("\tpxe.length: 0x%02x\r\n", pxe->structlength);
    NILO_TRACE_PRINTF ("\tpxe.structcksum: 0x%02x\r\n", pxe->structcksum);
    NILO_TRACE_PRINTF ("\tpxe.structrev: 0x%02x\r\n", pxe->structrev);

    NILO_TRACE_PRINTF ("\tpxe.undi: 0x%08x\r\n", pxe->undi);
    NILO_TRACE_PRINTF ("\tpxe.base: 0x%08x\r\n", pxe->base);

    NILO_TRACE_PRINTF ("\tpxe.entry16point: 0x%08x\r\n", pxe->entry16point);
    NILO_TRACE_PRINTF ("\tpxe.entry32point: 0x%08x\r\n", pxe->entry32point);
    NILO_TRACE_PRINTF ("\tpxe.statuscallout: 0x%08x\r\n", pxe->statuscallout);

    NILO_TRACE_PRINTF ("\tpxe.segdesccnt: 0x%02x\r\n", pxe->segdesccnt);

    NILO_TRACE_PRINTF ("\tpxe.firstselector: 0x%04x\r\n", pxe->firstselector);

    NILO_TRACE_PRINTF ("\tpxe.stack: 0x%08x\r\n", pxe->stack);
    NILO_TRACE_PRINTF ("\tpxe.undidata: 0x%08x\r\n", pxe->undidata);
    NILO_TRACE_PRINTF ("\tpxe.undicode: 0x%08x\r\n", pxe->undicode);
    NILO_TRACE_PRINTF ("\tpxe.undicodewrite: 0x%08x\r\n", pxe->undicodewrite);
    NILO_TRACE_PRINTF ("\tpxe.bc_data: 0x%08x\r\n", pxe->bc_data);
    NILO_TRACE_PRINTF ("\tpxe.bc_code: 0x%08x\r\n", pxe->bc_code);
    NILO_TRACE_PRINTF ("\tpxe.bc_codewrite: 0x%08x\r\n", pxe->bc_codewrite);
}

#if 0
#include "bootp.h"

/* dump the contents of a BOOTP packet */
void
dump_bootp_packet(bootp_t *bpr)
{
    unsigned char *p;
    int i;
    
    NILO_TRACE_FUNCTION;

    NILO_TRACE_MSG ("bootp.bp_opcode: 0x%x\r\n", bpr->bp_opcode);
    NILO_TRACE_MSG ("bootp.bp_htype: 0x%x\r\n", bpr->bp_htype);
    NILO_TRACE_MSG ("bootp.bp_hlen: 0x%x\r\n", bpr->bp_hlen);
    NILO_TRACE_MSG ("bootp.bp_hops: 0x%x\r\n", bpr->bp_hops);
    NILO_TRACE_MSG ("bootp.xid: 0x%04x\r\n", bpr->bp_xid);
    NILO_TRACE_MSG ("bootp.bp_secs: 0x%02x\r\n", bpr->bp_secs);
    NILO_TRACE_MSG ("bootp.bp_flags: 0x%02x\r\n", bpr->bp_flags);

    NILO_TRACE_MSG ("bootp.bp_ciaddr: 0x%04x, (%s)\r\n", bpr->bp_ciaddr.num,
                    _inet_ntoa(bpr->bp_ciaddr.num));
    NILO_TRACE_MSG ("bootp.bp_yiaddr: 0x%04x, (%s)\r\n", bpr->bp_yiaddr.num,
                    _inet_ntoa(bpr->bp_yiaddr.num));
    NILO_TRACE_MSG ("bootp.bp_siaddr: 0x%04x, (%s)\r\n", bpr->bp_siaddr.num,
                    _inet_ntoa(bpr->bp_siaddr.num));
    NILO_TRACE_MSG ("bootp.bp_giaddr: 0x%04x, (%s)\r\n", bpr->bp_giaddr.num,
                    _inet_ntoa(bpr->bp_giaddr.num));

    NILO_TRACE_MSG ("bootp.bp_chaddr: %s\r\n", hexify_buffer (bpr->bp_chaddr, BP_CHADDR_LEN/2));

    NILO_TRACE_MSG ("bootp.bp_sname: %s\r\n", (*bpr->bp_sname) ? (char *)bpr->bp_chaddr: "none yet");
    NILO_TRACE_MSG ("bootp.bp_file: %s\r\n", (*bpr->bp_file) ? bpr->bp_file: "none yet");

#ifdef NILO_PXE_COMP
    union {
        u_int8_t d[BP_VEND_LEN];        /* vendor-specific area */
        struct {
            u_int8_t  magic[4];         /* magic number */
            u_int32_t flags;            /* flags/opcodes etc */
            u_int8_t  pad[56];          /* padding chars */
        } v;
    } bp_options;
#else
    NILO_TRACE_MSG ("bootp.bp_vend:");
    for (p = bpr->bp_options; p < &bpr->bp_options[200] && *p != TAG_END; ++p)
        NILO_TRACE_PRINTF (" 0x%02x", *p);
    NILO_TRACE_PRINTF ("\r\n");
#endif
    NILO_TRACE_FUNCTION_RETURN;
}
#endif

/* dump the contents of a DHCP packet */
void
dump_dhcp_packet(dhcp_t *dhcp_pkt)
{
    u_int8_t *p;
    char s[DHCP_SNAME_LEN];
    int i, j, len, endtag, num;
    
    NILO_TRACE_FUNCTION;

    NILO_TRACE_MSG ("Dumping a DHCP packet.\r\n");
    NILO_TRACE_PRINTF ("\tdhcp.bp_opcode: 0x%x\r\n", dhcp_pkt->bp_opcode);
    NILO_TRACE_PRINTF ("\tdhcp.bp_htype: 0x%x\r\n", dhcp_pkt->bp_htype);
    NILO_TRACE_PRINTF ("\tdhcp.bp_hlen: 0x%x\r\n", dhcp_pkt->bp_hlen);
    NILO_TRACE_PRINTF ("\tdhcp.bp_hops: 0x%x\r\n", dhcp_pkt->bp_hops);
    NILO_TRACE_PRINTF ("\tdhcp.xid: 0x%04x\r\n", dhcp_pkt->bp_xid);
    NILO_TRACE_PRINTF ("\tdhcp.bp_secs: 0x%02x\r\n", dhcp_pkt->bp_secs);
    NILO_TRACE_PRINTF ("\tdhcp.bp_flags: 0x%02x\r\n", dhcp_pkt->bp_flags);

    NILO_TRACE_PRINTF ("\tdhcp.bp_ciaddr: 0x%04x, (%s)\r\n",
                       dhcp_pkt->bp_ciaddr.num,
                    _inet_ntoa(dhcp_pkt->bp_ciaddr.num));
    NILO_TRACE_PRINTF ("\tdhcp.bp_yiaddr: 0x%04x, (%s)\r\n",
                       dhcp_pkt->bp_yiaddr.num,
                    _inet_ntoa(dhcp_pkt->bp_yiaddr.num));
    NILO_TRACE_PRINTF ("\tdhcp.bp_siaddr: 0x%04x, (%s)\r\n",
                       dhcp_pkt->bp_siaddr.num,
                    _inet_ntoa(dhcp_pkt->bp_siaddr.num));
    NILO_TRACE_PRINTF ("\tdhcp.bp_giaddr: 0x%04x, (%s)\r\n",
                       dhcp_pkt->bp_giaddr.num,
                    _inet_ntoa(dhcp_pkt->bp_giaddr.num));

    NILO_TRACE_PRINTF ("\tdhcp.bp_chaddr: %s\r\n",
                       hexify_buffer (dhcp_pkt->bp_chaddr, 
                                      (DHCP_CHADDR_LEN/2)-2));

    NILO_TRACE_PRINTF ("\tdhcp.bp_sname: %s\r\n",
                       (*dhcp_pkt->bp_sname && isascii(*dhcp_pkt->bp_sname))
                       ? (char *)dhcp_pkt->bp_sname: "none yet");
    NILO_TRACE_PRINTF ("\tdhcp.bp_file: %s\r\n", (*dhcp_pkt->bp_file)
                       ? dhcp_pkt->bp_file: "none yet");

    NILO_TRACE_PRINTF ("\tdhcp.bp_options:");

    /* Look for the last 0xff in the option field, cause it's the
       end tag */
    endtag = DHCP_MAX_OPT_SIZE;
    while (dhcp_pkt->bp_options[endtag--] != TAG_END) {    }

    for (p = dhcp_pkt->bp_options; p < &dhcp_pkt->bp_options[DHCP_VEND_LEN] && p < (dhcp_pkt->bp_options + endtag); ++p)
        NILO_TRACE_PRINTF (" 0x%02x", *p);
    NILO_TRACE_PRINTF (" 0x%02x", *p);  /* print the end tag */
    NILO_TRACE_PRINTF ("\r\n");
    
    p = dhcp_pkt->bp_options;
    i = 4;                              /* ignore the vendor magic */

    while (p[i] != TAG_END && p[i] != 0) {
        NILO_TRACE_PRINTF ("\t\toption %s (0x%02x): ", optnames[p[i-1]], p[i++]);
        len = p[i++];
        switch (p[i-2]) {
        case DHCP_PXE1_CMD_LINE:
        case DHCP_PXE_CMD_LINE:
        case DHCP_DOMAINNAME:
        case DHCP_BOOTFILE:
        case DHCP_CLIENT_IDENTIFIER:    /* option 61 */
            memcpy (s, (char *)p+i, len);
            s[len] = '\0';
            NILO_TRACE_PRINTF ("%s\r\n", s);
            break;
        case DHCP_REQUESTED_IP_ADDR:
        case DHCP_SUBNET_MASK:
        case DHCP_ROUTERS_ON_SUBNET:
        case DHCP_SERVER_IDENTIFIER:
        case DHCP_BROADCAST_ADDR:
            memcpy (&num, p+i, len);
            NILO_TRACE_PRINTF ("%s\r\n", _inet_ntoa(num));
            break;
        case DHCP_MESSAGE_TYPE:
            NILO_TRACE_PRINTF ("type %s (0x%02x): \r\n", msgtypes[p[i]], p[i]);
            break;
        case DHCP_PLATFORMID:           /* option 97 */
            memcpy (s, (char *)p+i+1, len-1);
            s[len-1] = '\0';
            NILO_TRACE_PRINTF ("Platform Type is %d, ID is %s\r\n",
                               p[i], s); 
            break;
        case DHCP_NICIF:                /* option 94 */
            NILO_TRACE_PRINTF ("NIC Type 0x%1x, ", p[i]); 
            NILO_TRACE_PRINTF ("Version %d.%d\r\n", p[i+1], p[i+2]); 
            break;
        case DHCP_SYSARCH:              /* option 93 */
            NILO_TRACE_PRINTF ("The System Arch is %s\r\n", archtypes[p[i+1]]);
            break;
        case DHCP_MAXMSGSIZE:           /* option 57 */
            memcpy (&num, p+i, len);
            NILO_TRACE_PRINTF ("Max Msg length 0x%02x\r\n", num);
            break;
        case DHCP_CLASS_IDENTIFIER:     /* option 60 */
            memcpy (s, (char *)p+i, len);
            s[len] = '\0';
            NILO_TRACE_PRINTF ("%s\r\n", s);
            break;
        default:
            for (j=0; j<len; j++)
                NILO_TRACE_PRINTF (" 0x%02x", p[i+j]);
            NILO_TRACE_PRINTF ("\r\n");
        }
        i+=len;
    }

    NILO_TRACE_FUNCTION_RETURN;
}


/* dump the contents of a TFTP packet */
void
dump_tftp_packet(tftp_t *pkt)
{
    unsigned char *p;
    int i;
    
    NILO_TRACE_FUNCTION;

    if (pkt == 0) {
        NILO_TRACE_MSG ("ERROR: tftp structure is NULL\r\n");
        return;
    }
    NILO_TRACE_MSG ("Dumping a TFTP packet..\r\n");

    switch (ntohs(pkt->opcode)) {
    case TFTP_RRQ:
        NILO_TRACE_PRINTF ("\ttftp.opcode: RRQ\r\n");
        NILO_TRACE_PRINTF ("\ttftp.u.data: %s\r\n", pkt->u.rrq);
        NILO_TRACE_PRINTF ("\ttftp mode is: %s\r\n",
                           (pkt->u.rrq+strlen(pkt->u.rrq)+1));
        break;
    case TFTP_WRQ:
        NILO_TRACE_PRINTF ("\ttftp.opcode: WRQ\r\n");
        NILO_TRACE_PRINTF ("\ttftp.u.data: %s\r\n", pkt->u.rrq);
        NILO_TRACE_PRINTF ("\ttftp mode is: %s\r\n",
                           (pkt->u.rrq+strlen(pkt->u.rrq)+1));
        break;
    case TFTP_ACK:
        NILO_TRACE_PRINTF ("\ttftp.opcode: ACK\r\n");
        NILO_TRACE_PRINTF ("\ttftp.u.ack.block: 0x%02x\r\n",
                           ntohs(pkt->u.ack.block));
        break;
    case TFTP_OACK:
        NILO_TRACE_PRINTF ("\ttftp.opcode: OACK\r\n");
        NILO_TRACE_MSG ("FIXME: we don't handle OACK packets yet.\n");
        break;
    case TFTP_DATA:
        NILO_TRACE_PRINTF ("\ttftp.opcode: DATA\n");
        NILO_TRACE_PRINTF ("\ttftp.data.block: 0x%02x\r\n",
                           ntohs(pkt->u.data.block));
        NILO_TRACE_PRINTF ("\ttftp.u.data: %s lots more...\r\n",
                           hexify_buffer(pkt->u.data.download, 30));

    }

    NILO_TRACE_FUNCTION_RETURN;
}

void
dump_ip_packet (struct iphdr *ip)
{
    u_int32_t num;

    NILO_TRACE_FUNCTION;

    NILO_TRACE_MSG ("Dumping an IP packet..\r\n");
    NILO_TRACE_PRINTF ("\tip.version: 0x%0x\r\n", ip->verhdrlen >> 4);
    NILO_TRACE_PRINTF ("\tip.hdrlen: 0x%0x\r\n", ip->verhdrlen & 0xf);
    NILO_TRACE_PRINTF ("\tip.service: 0x%02x\r\n", ip->service);
    NILO_TRACE_PRINTF ("\tip.len: 0x%04x\r\n", ntohs(ip->len));
    NILO_TRACE_PRINTF ("\tip.ident: 0x%04x\r\n", ntohs(ip->ident));
    NILO_TRACE_PRINTF ("\tip.frags: 0x%04x\r\n", ntohs(ip->frags));
    NILO_TRACE_PRINTF ("\tip.ttl: 0x%02x\r\n", ip->ttl);
    NILO_TRACE_PRINTF ("\tip.protocol: 0x%02x\r\n", ip->protocol);
    NILO_TRACE_PRINTF ("\tip.chksum: 0x%04x\r\n", ntohl(ip->chksum));
    memcpy (&num, ip->src, 4);
    NILO_TRACE_PRINTF ("\tip.src: 0x%04x (%s)\r\n", ntohl(num), _inet_ntoa(num));
    memcpy (&num, ip->dest, 4);
    NILO_TRACE_PRINTF ("\tip.dest: 0x%04x (%s)\r\n", ntohl(num), _inet_ntoa(num));
    
    NILO_TRACE_FUNCTION_RETURN;
}

void
dump_udp_packet (struct udphdr *udp)
{
    NILO_TRACE_FUNCTION;

    NILO_TRACE_MSG ("\tDumping a UDP packet..\r\n");

    NILO_TRACE_PRINTF ("\tudp.src:  0x%04x\r\n", ntohs(udp->src));
    NILO_TRACE_PRINTF ("\tudp.dest: 0x%04x\r\n", ntohs(udp->dest));
    NILO_TRACE_PRINTF ("\tudp.len: 0x%04x\r\n", ntohs(udp->len));
    NILO_TRACE_PRINTF ("\tudp.chksum: 0x%04x\r\n", udp->chksum);

    NILO_TRACE_FUNCTION_RETURN;
}

void
dump_arp_table(arptable_t *arptbl)
{
    int i;

    NILO_TRACE_FUNCTION;

    NILO_TRACE_MSG ("Dumping the ARP table\r\n");
    for (i=0; i<MAX_ARP; i++) {
        NILO_TRACE_PRINTF ("\t\t%s", arpentries[i]);
        NILO_TRACE_PRINTF ("\t%s", _inet_ntoa(arptbl[i].ipaddr.num));
        if (arptbl[i].ipaddr.num == 0)
            NILO_TRACE_PRINTF ("\t");
        NILO_TRACE_PRINTF ("\t\t%s\r\n", hexify_buffer (arptbl[i].hwaddr, 6));
    }

    NILO_TRACE_FUNCTION_RETURN;
}

/* Dump an ARP packet */
void
dump_arp_packet (struct arprequest *pkt)
{
    u_int32_t temp;

    NILO_TRACE_FUNCTION;

    NILO_TRACE_MSG ("Dumping an ARP packet...\r\n");

    NILO_TRACE_PRINTF ("\t\tarprequest.hwtype: 0x%02x\r\n", 
                           ntohs(pkt->hwtype));
    
    if (ntohs(pkt->protocol) == ETHERTYPE_IP)
        NILO_TRACE_PRINTF ("\t\tarprequest.protocol: 0x%02x\r\n",
                           ntohs(pkt->protocol));
    else
        NILO_TRACE_PRINTF ("\t\tBAD protocol type, 0x%02x!\r\n",
                           ntohs(pkt->protocol));
   
    if (pkt->hwlen == 6)
        NILO_TRACE_PRINTF ("\t\tarprequest.hwlen: 0x%01x\r\n",
                           pkt->hwlen);
    else
        NILO_TRACE_PRINTF ("\t\tBAD hardware address length!\r\n",
                           pkt->hwlen);
   
    if (pkt->protolen == 4)
        NILO_TRACE_PRINTF ("\t\tarprequest.protolen: 0x%01x\r\n",
                           pkt->protolen);
    else
        NILO_TRACE_PRINTF ("BAD protocol length, 0x%01x!\r\n",
                           pkt->protolen);
       
    NILO_TRACE_PRINTF ("\t\tarprequest.opcode: %s\r\n",
                       arpops[ntohs(pkt->opcode)]);

    NILO_TRACE_PRINTF ("\t\tarprequest.shwaddr: %s\r\n",
                       hexify_buffer (pkt->shwaddr, 6));

    memcpy(&temp, pkt->sipaddr, sizeof(temp));
    NILO_TRACE_PRINTF ("\t\tarprequest.sipaddr: 0x%04x, (%s)\r\n",
                       temp, _inet_ntoa(temp));
//                       pkt->sipaddr, _inet_ntoa(pkt->sipaddr));

    NILO_TRACE_PRINTF ("\t\tarprequest.thwaddr: %s\r\n",
                       hexify_buffer (pkt->thwaddr, 6));
    memcpy(&temp, pkt->tipaddr, sizeof(temp));
    NILO_TRACE_PRINTF ("\t\tarprequest.tipaddr: 0x%04x, (%s)\r\n",
                       temp, _inet_ntoa(temp));

    NILO_TRACE_FUNCTION_RETURN;
}

/* This reads in a tagged image boot file, and decodes the header fields */
void
dump_tagged (union infoblock *dirblk)
{
    int i, x, vlen, nlen, flags;
    char *vinfo;
    int infile_fd;
    struct imgheader *imghead;
    struct segheader *seghead, *foo;
    struct segheader sh = { 0, 0, 0, 0, 0L, 0L, 0L };

    NILO_TRACE_MSG ("Dumping a tagged image file header block...\r\n");

    if (dirblk->i.magic != TAG_MAGIC || dirblk->s[255] != 0xAA55) {
        NILO_TRACE_PRINTF ("ERROR: Tagged file magic number is incorrect.\r\n");
        exit(0);
    } else {
        NILO_TRACE_PRINTF ("\tThis is a tagged bootfile.\r\n");
    }

    imghead = (struct imgheader *)dirblk;
    NILO_TRACE_PRINTF ("\tThe Vendor length is 0x%02x bytes\n\tnon-vendor length is 0x%02x bytes.\r\n"
             , imghead->vlen*4, imghead->nlen*4);
    vinfo = (char *)dirblk + sizeof(struct imgheader);
    
    NILO_TRACE_PRINTF ("\tImage %s return to NILO after execution.\r\n",
             (imghead->bootret) ? "can" : "can't");
    NILO_TRACE_PRINTF ("\tLocation: 0x%04hx:0x%04hx\r\n\tExecaddr: 0x%04hx:0x%04hx\r\n",
           imghead->location.offset, imghead->location.segment,
           imghead->execaddr.offset, imghead->execaddr.segment);
    if (imghead->vlen > 0)
        NILO_TRACE_PRINTF ("\tVendor info: %s\r\n", vinfo);

    /* get the first segment */
    seghead = (struct segheader *)(((char *)dirblk) + (imghead->vlen*4)
                                   + (imghead->nlen*4));
    for (i=1; i<=10; i++) {
        NILO_TRACE_PRINTF ("\t\n\tSegment header #%d.\r\n", i);
        if (i>0)
            seghead = (struct segheader *)(((char *)dirblk) + (seghead->vlen*4)
                                           + (seghead->nlen*4)
                                           + (i*(sizeof(struct segheader))));
        if (seghead->vlen)
            vinfo = (char *)(seghead + sizeof(struct segheader));

        NILO_TRACE_PRINTF ("\tThe Segment Header Vendor length is 0x%02x bytes.\r\n",
                 seghead->vlen*4);
        NILO_TRACE_PRINTF ("\tThe Segment Header non-vendor length is 0x%02x bytes.\r\n",
                 seghead->nlen*4);
        NILO_TRACE_PRINTF ("\tThe Segment Vendor Tag is 0x%02x.\r\n", seghead->vendorinfo);
        /* FIXME: These are likely to be Linux specific */
        switch (seghead->vendorinfo) {
        case VTAG_BOOT_LOADER:
            NILO_TRACE_PRINTF ("\t\tLinux primary boot loader.\r\n");
            break;
        case VTAG_CMD_LINE:
            NILO_TRACE_PRINTF ("\t\tLinux command line.\r\n");
            break;
        case VTAG_FLOPPY_SECTOR:
            NILO_TRACE_PRINTF ("\t\tLinux floppy boot sector.\r\n");
            break;
        case VTAG_KERNEL_SETUP:
            NILO_TRACE_PRINTF ("\t\tLinux kernel setup.\r\n");
            break;
        case VTAG_KERNEL_IMAGE:
            NILO_TRACE_PRINTF ("\t\tLinux kernel image.\r\n");
            break;
        case VTAG_RAMDISK:
            NILO_TRACE_PRINTF ("\t\tLinux ramdisk image.\r\n");
            break;
        }

        NILO_TRACE_PRINTF ("\tThe Segment flags are 0x%01x.\r\n", seghead->flags);
        switch (seghead->flags) {
        case LOAD_ABSOLUTE:
            NILO_TRACE_PRINTF ("\t\tLoading to an absolute 32 bit address.\r\n");
            break;
        case LOAD_SUB_ADDR:
            NILO_TRACE_PRINTF ("\t\tLoading to the last writeable location in memory.\r\n");
            break;
        case LOAD_ADD_ADDR:
            NILO_TRACE_PRINTF ("\t\tLoad address is 1 past the last image.\r\n");
            break;
        case LOAD_SUB_START:
            NILO_TRACE_PRINTF ("\t\tLoad address is subtracted from the start of the last image.\r\n");
            break;
        }
        NILO_TRACE_PRINTF ("\tThe Segment endmark %s set.\r\n", (seghead->endmark) ? "is" : "isn't");
        NILO_TRACE_PRINTF ("\tThe load address for this segment is 0x%08x.\r\n", seghead->loadaddr);
        NILO_TRACE_PRINTF ("\tThe image length for this segment is 0x%08x bytes.\r\n", seghead->imglength);
        NILO_TRACE_PRINTF ("\tThe memory length for this segment is 0x%08x bytes.\r\n", seghead->memlength);
        
         if (seghead->vlen > 0) {
            /* if we're a ramdisk image */
            if (seghead->vendorinfo == 0x15) {
                switch (*vinfo) {
                case RD_AUTO:
                    NILO_TRACE_PRINTF ("\t\tVendor info: RD: auto positioning.\r\n");
                    break;
                case RD_EOM:
                    NILO_TRACE_PRINTF ("\t\tVendor info: positioned by Boot-Rom.\r\n");
                    break;
                case RD_FIXED:
                    NILO_TRACE_PRINTF ("\t\tVendor info: RD: fixed memory location.\r\n");
                    break;
                }
            }
        } else {
            NILO_TRACE_PRINTF ("\tThere is no vendor info.\r\n");
        }

        if (seghead->endmark)
            break;
    } /* for i */

}

