/* Nilo - A Network boot loader.

   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the PXE data structures */
#include "pxe.h"

/* Define the PXE UDP data structures */
#include "pxe_udp.h"

/* Define the tracing macros */
#include "trace.h"

/* Define the data structure dumping functions for debug use */
#include "dump.h"

#include "nilo.h"

/* Define the standard BSD TFTP data types */
#include <oskit/freebsd/sys/types.h>
#include <net/if.h>
#include <arpa/tftp.h>
#include <net/if_arp.h>

#include <stdlib.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include "tagged.h"
#include "util.h"
#include <oskit/x86/far_ptr.h>
#include <oskit/x86/base_vm.h>

#include "pxenv.h"
#include "pxe.h"
#include "pxe_undi.h"
#include "pxe_tftp.h"
#include "dhcp.h"

#ifdef NILO_USE_MULTIBOOT
extern oskit_addr_t reboot_stub_loc;
#include <oskit/x86/pc/base_multiboot.h> /* boot_info */
#include <assert.h>
#else
#endif /* NILO_USE_MULTIBOOT */
int oskit_bootargc;
char **oskit_bootargv;
#undef	SIMULATE_BLOCK_LOSS

#include <oskit/gdb.h>
#include <oskit/x86/base_gdt.h>
#include <oskit/x86/i16.h> /* boot_info */

void copy_base_gdt(void);

extern u_int32_t jiffies;

struct client_data *client_info;
static int32_t timeout;
static u_int32_t xid;

pxe_t *pxe_data_struct;
pxenv_t *pxenv_data_struct;

int32_t seed;	/* for random number generator */

extern char do_boot[], do_boot_end[];

extern void jumpto(segoff16_t execaddr, segoff16_t location,
             dhcp_t *bootdata);

// static void dhcp_callback (int32_t snaplen, dhcp_t *dhcp_ptr);
// static void process_dhcp_reply (dhcp_t *dhcp_ptr);

/**************************************************************************
RFC951_SLEEP - sleep for expotentially longer times
**************************************************************************/
int32_t
rfc951_sleep(int32_t exp)
{
    int32_t q,z,tmo;

    NILO_TRACE_FUNCTION;

    /* simplified version of the LCG given in Bruce Scheier's
       "Applied Cryptography" */
    q = seed/53668;
    if ((seed = 40014*(seed-53668*q) - 12211*q) < 0) seed += 2147483563l;
    /* compute mask */
    for (tmo = 63; tmo <= 60*1008 && --exp > 0; tmo = 2*tmo+1);
    return (tmo & seed);

    NILO_TRACE_FUNCTION_RETURN;
}

void
jump_to_code(segoff16_t exec, segoff16_t loc, dhcp_t *bootdata)
{
    oskit_addr_t entry, load_addr;
    oskit_size_t stub_size;
    void *stub;
    struct far_pointer_32 ptr;
    oskit_addr_t copy_source;
    oskit_size_t copy_size;
    
    NILO_TRACE_FUNCTION;
#ifdef NILO_USE_TRACING
    NILO_TRACE_MSG ("Jumping to 0x%08x\r\n", client_info->execaddr);
#else  /* FIXME: this is just temporary */
    printf ("Jumping to 0x%08x\r\n", client_info->execaddr);
#endif
    cli();

    /*
     * All of the data structures that are important to the kernel,
     * are in memory guaranteed not to conflict with the kernel image
     * or with anything else.
     * However, the kernel image itself is not in its final position,
     * because that might be right on top of us,
     * or on top of anything we allocated before reserving the
     * kernel image region.
     * Therefore, we must allocate (non-conflicting) memory for
     * a small stub to copy the kernel image to its final position
     * and invoke it.
     */
    stub_size = do_boot_end - do_boot;
    stub = mustmalloc(stub_size);
    memcpy(stub, do_boot, stub_size);
    
    ptr.seg = LINEAR_CS;
    ptr.ofs = kvtophys(stub);
    
    /*
     * The kernel image source and destination regions may overlap,
     * so we may have to copy either forwards or backwards.
     *
     * XXX the "mr" constraint for "ptr" should be "m" but GCC
     * says "inconsistent operand constraints."
     *
     * Be careful what you put as the values for the registers in
     * the asm's below.
     * If you have too many indirections and whatnot then GCC
     * will apparently run out of registers and won't be able to
     * set, for example, %ebx to the desired value.
     * The GCC chokeage looked like this for me:
     *
     *      fixed or forbidden register was spilled.
     *      This may be due to a compiler bug or to impossible asm
     *
     * That is why I have load_addr and entry be local vars instead of
     * just putting the expressions in the asm.
     */
//    copy_source = kvtophys(ki->ki_imgaddr);
    copy_source = (u_int32_t)client_info->location.segment << 4 
        + client_info->location.offset;
    copy_size = client_info->imglen;

    entry = (u_int32_t)client_info->execaddr.segment << 4 
        + client_info->execaddr.offset;
    load_addr = (u_int32_t)client_info->location.segment << 4 
        + client_info->location.offset;
    
#if 0
    dprintf("ki->ki_mbinfo = %p-%p\n",
            kvtophys(ki->ki_mbinfo),
            kvtophys(ki->ki_mbinfo) + sizeof(ki->ki_mbinfo));
#endif
    NILO_TRACE_MSG("stub = %p-%p\n", stub, stub + stub_size);
    
    if (copy_source > load_addr) {
        asm volatile("
            cld
            ljmp    %0
            " : /* no outputs */
              : "mr" (ptr),
                "a" (entry),  // a register
                "S" (copy_source),  // si register
                "D" (load_addr),   // di registers
                "c" (copy_size),    // c register
                "b" (kvtophys(&client_info)),      // b register
                "d" (LINEAR_DS));   // d register
    } else {
        asm volatile("
            std
            ljmp    %0
            " : /* no outputs */
              : "mr" (ptr),
                "a" (entry),
                "S" (copy_source + copy_size - 1),
                "D" (load_addr + copy_size - 1),
                "c" (copy_size),
                "b" (kvtophys(&client_info)),
                "d" (LINEAR_DS));
        }

    NILO_TRACE_FUNCTION_RETURN;
}

#ifdef NILO_USE_MULTIBOOT
/* We grabbed these from OSKit's netboot, cause these process the
   console options so we can use the serial port */

/*
 * Figure out if -h (serial) or -f (fast serial) are in the oskit_bootargv
 * and put them in the default_args list.
 */
static void
init_default_args(int bootargc, char *bootargv[], char *default_args[])
{
    char **p;
    int i;
    
    NILO_TRACE_FUNCTION;

    p = &default_args[0];
    for (i = 0; i < bootargc; i++)
#define NDEFAULT_ARGS 2
        if (strcmp(bootargv[i], "-h") == 0 ||
            strcmp(bootargv[i], "-f") == 0)
            *p++ = bootargv[i];
    *p = NULL;

    NILO_TRACE_FUNCTION_RETURN;
}

/*
 * Build a command line from the program name, the default_args,
 * what they specified, and our -retaddr option.
 *
 * The command line is in a special format understood by the oskit
 * multiboot startup code (see base_multiboot_init_cmdline.c).
 * Basically, the format separates booting-options and environment
 * variable settings from normal args to pass to main().  The format is like
 *      progname [<booting-options and foo=bar> --] <args to main>
 * For example
 *      kernel DISPLAY=host:0 -d -- -rf foo bar
 */
static char *
build_cmdline(char *progname, char *default_args[NDEFAULT_ARGS], char *input)
{
    char *toks[input? strlen(input) : 0];
    char *tok;
    int ntoks;
    int i;
    int havesep;
    int cllen;
    char *cmdline;
    NILO_TRACE_FUNCTION;

    /* Tokenize their input. */
    ntoks = 0;
    if (input)
        for (tok = strtok(input, " \t"); tok; tok = strtok(0, " \t"))
            toks[ntoks++] = tok;
    
    /* Figure out if they specified a "--". */
    havesep = 0;
    for (i = 0; i < ntoks; i++)
        if (strcmp(toks[i], "--") == 0) {
            havesep = 1;
            break;
        }
    
    /*
     * Toggle off any bootopts that they specified.
     * We only have to do this if there is a --, since otherwise
     * they didn't specify any bootopts.
     */
    if (havesep) {
        for (i = 0; i < ntoks; i++) {
            int j;
            
            if (strcmp(toks[i], "--") == 0)
                break;
            for (j = 0; j < NDEFAULT_ARGS && default_args[j]; j++)
                if (strcmp(toks[i], default_args[j]) == 0) {
                    toks[i] = 0;
                    default_args[j] = 0;
                }
        }
    }
    
    /* Figure out how big the command-line will be and allocate it. */
    cllen = strlen(progname) + 1;
    cllen += strlen("-retaddr 0x00ff00ff") + 1;
    for (i = 0; i < NDEFAULT_ARGS; i++)
        if (default_args[i])
            cllen += strlen(default_args[i]) + 1;
    if (! havesep)
        cllen += strlen("--") + 1;
    for (i = 0; i < ntoks; i++)
        if (toks[i])
            cllen += strlen(toks[i]) + 1;
    cllen += 1;                     /* for NUL */
    
    cmdline = mustcalloc(cllen,1);
    assert(cmdline);
    
    /*
     * Now fill it in.
     */
    strcpy(cmdline, progname);
    strcat(cmdline, " ");
    
//    sprintf(cmdline + strlen(progname) + 1, "-retaddr 0x%08x",
//            reboot_stub_loc);
//    strcat(cmdline, " ");
    
    for (i = 0; i < NDEFAULT_ARGS; i++)
        if (default_args[i]) {
            strcat(cmdline, default_args[i]);
            strcat(cmdline, " ");
        }
    
    if (! havesep) {
        strcat(cmdline, "--");
        strcat(cmdline, " ");
    }
    
    for (i = 0; i < ntoks; i++)
        if (toks[i]) {
            strcat(cmdline, toks[i]);
            strcat(cmdline, " ");
        }
    
    NILO_TRACE_FUNCTION_RETURN;
    return cmdline;
}
#endif /* NILO_USE_MULTIBOOT */

int 
main(int argc, char **argv)
{
    int32_t retries;
    struct in_addr addr;
    pxenv_undi_initialize_t undiinit;
    pxenv_udp_open_t udpopen;
    pxenv_tftp_read_file_t tftpfile;

#ifdef NILO_USE_MULTIBOOT
    char *default_args[NDEFAULT_ARGS+1] = {0};
#endif /* NILO_USE_MULTIBOOT */

    NILO_TRACE_FUNCTION;
    
    NILO_TRACE_PRINTF ("NILO dated " VERSION "\r\n");
    init_system();

#ifdef NILO_USE_MULTIBOOT
    init_default_args(oskit_bootargc, oskit_bootargv, default_args);
#endif /* NILO_USE_MULTIBOOT */

    if ((client_info = malloc(sizeof(*client_info))) == 0) {
        NILO_TRACE_MSG("Malloc failed\r\n");
    }

    if (CALL_PXE_API (PXENV_UNDI_INITIALIZE, &undiinit) == PXENV_EXIT_FAILURE) {
        panic ("Couldn't initialize the UNDI system\r\n");
        return (PXENV_EXIT_FAILURE);
    }

    /* Open the UDP connection. This also does a DHCP session to get
       the boot parameters. */
    if (CALL_PXE_API (PXENV_UDP_OPEN, &udpopen) == PXENV_EXIT_FAILURE) {
        panic ("Couldn't open the UDP connection.\r\n");
        return (PXENV_EXIT_FAILURE);
    }

    /* The IP address of the TFTP server in Big Endian format. */
    tftpfile.server_ip.num = client_info->siaddr.num;
    /* The IP address of the TFTP gateway in Big Endian format. */
    tftpfile.gw_ip.num = client_info->yiaddr.num;
    /* The size of the receive buffer in bytes. */
    tftpfile.buffersize = TFTP_DEFAULTSIZE_PACKET;
    /* The timeout value in seconds used for receiving data. If zero,
       the default of TFTP timeout is used. */
    tftpfile.tftp_open_timeout = 0;
    /* The maximum wait delay time in seconds between ACK of a fast
       packet and a new TFTP open. */
    tftpfile.tftp_reopen_delay = 10;
    /* The bootfile name. */
    strcpy (tftpfile.filename, client_info->file);
    /* the port the client listens to for a multicast. */
    tftpfile.tftp_clnt_port = TFTP_CLIENT_PORT;
    client_info->tftp_client_port = TFTP_CLIENT_PORT;
    /* The port the server sends to for a multicast. */
    tftpfile.tftp_srv_port = TFTP_SERVER_PORT;

    /* Load a file via a (closed) TFTP connection */
    if (CALL_PXE_API (PXENV_TFTP_READ_FILE, &tftpfile) == PXENV_EXIT_FAILURE) {
        panic ("Couldn't open the TFTP connection, and load a file.\r\n");
        return (PXENV_EXIT_FAILURE);
    }

    /* Start the TFTP transfer */
    //    client_info->tftp_handler(client_info->siaddr.num, TFTP_RRQ,
    //                      client_info->file, client_info->tftp_blockno);

    /* execute the new image */
    /* FIXME: Was jump_to_code */
    NILO_TRACE_MSG ("location.offset 0x%04x, location.segment 0x%04x\r\n",
                    client_info->location.offset, client_info->location.segment);
    NILO_TRACE_MSG ("execaddr.offset 0x%04x, execaddr.segment 0x%04x\r\n",
                    client_info->execaddr.offset, client_info->execaddr.segment);
    NILO_TRACE_MSG ("DHCP data record 0x%08x\r\n", client_info);

    jump_to_code (client_info->execaddr,
           client_info->location, (dhcp_t *)client_info);

#if 0
    jumpto(client_info->execaddr,
           client_info->location, (dhcp_t *)client_info);
#endif

    NILO_TRACE_FUNCTION_RETURN;

    return (0);
}

