/* 
   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the tracing macros */
#include "trace.h"

#include "nilo.h"

#include "pxe.h"
#include "pxenv.h"
#include "pxe_udp.h"
#include "pxe_tftp.h"
#include "pxe_undi.h"
#include "pxe_preboot.h"

#include <oskit/dev/dev.h>
#include <oskit/dev/net.h>
#include <oskit/dev/error.h>
#include <oskit/net/ether.h>
#include <oskit/dev/ethernet.h>
#include <oskit/dev/linux.h>
#include <oskit/dev/device.h>
#include <oskit/io/netio.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "arp.h"
#include "udp.h"
#include "ip.h"

#include <oskit/c/string.h>

void bumpticks(void);

extern pxe_t *pxe_data_struct;
extern pxenv_t *pxenv_data_struct;

/* These symbols are defined in the linker script for LD */
extern unsigned char *_etext;
extern unsigned char *_edata;
extern unsigned char *_end;

#if 0
u_int32_t *api_table[] = {
    (u_int32_t *)pxenv_get_binl_info,
    (u_int32_t *)udp_open,
    (u_int32_t *)udp_write,
    (u_int32_t *)udp_read,
    (u_int32_t *)udp_close,
    (u_int32_t *)tftp_get_fsize,
    (u_int32_t *)tftp_open,
    (u_int32_t *)tftp_read,
    (u_int32_t *)tftp_close,
    (u_int32_t *)tftpread_file,
    (u_int32_t *)undi_shutdown,
    (u_int32_t *)start_undi,
    (u_int32_t *)undi_startup,
    (u_int32_t *)undi_initialize,
    (u_int32_t *)undi_open,
    (u_int32_t *)undi_transmit,
    (u_int32_t *)undi_set_mcast_addr,
    (u_int32_t *)undi_set_station_addr,
    (u_int32_t *)undi_set_packet_filter,
    (u_int32_t *)undi_get_information,
    (u_int32_t *)undi_get_statistics,
    (u_int32_t *)undi_clear_statistics,
    (u_int32_t *)undi_initiate_diags,
    (u_int32_t *)undi_force_interrupt,
    (u_int32_t *)undi_get_mcast_addr,
    (u_int32_t *)undi_get_nic_type,
    (u_int32_t *)undi_get_ndis_info,
    (u_int32_t *)undi_isr,
    (u_int32_t *)undi_close,
    (u_int32_t *)undi_reset_nic,
    (u_int32_t *)undi_shutdown,
    (u_int32_t *)undi_cleanup,
    (u_int32_t *)stop_undi
};
#endif

bc_romid_t nilo_bc_romid = {
    BC_ROMID_SIG,               /* Structure signature is not NULL */
				/* terminated. */
    0x12,                       /* Length of this structure in */
				/* bytes. */
    0,                          /* Use to make byte checksum of this */
				/* structure == zero. */
    BC_ROMID_REV,               /* Structure format revision number. */
    PXE_REV,                    /* api revision number stored in */
       				/* Intel order. */
       				/* Revision 2.1.0 == 0x00, 0x01, 0x02 */
    0,                          /* Offset of base-code loader routine */
       				/* in the option ROM image. */
    0,                          /* Minimum stack segment size, in */
       				/* bytes, needed to load and run the */
       				/* base-code. */
    0,                          /* Base-code runtime code and data */
    0                           /* segment sizes. */
};

/* page 85 */
undi_romid_t nilo_undi_romid = {
    UNDI_ROMID_SIG,             /* Structure signature. */
    0x16,                       /* Length of this structure in bytes. */
    0,                          /* Use to make byte checksum of this */
				/* structure == zero. */
    UNDI_ROMID_REV,             /* Structure format revision number. */
    PXE_REV,                    /* API revision number stored in */
       				/* Intel order. */
       				/* Revision 2.1.0 == 0x00, 0x01, 0x02 */
    0,                          /* Offset of UNDI loader routine */
				/* in the option ROM image. */
    0,                          /* Minimum stack segment size, in */
				/* bytes, needed to load and run UNDI. */
    0,                          /* UNDI runtime code and data */
    0,                          /* segment sizes. */
    0                           /* 'ISAR', 'EISA', 'PCIR', 'PCCR' */
};

/* This is our default !PXE structure as used by all new PXE compliant
   code. All older applications are requested to be modified to use this
   structure rather than the older PXENV+ one. */
pxe_t *pxe_data_struct;
#if 1
pxe_t nilo_pxe = {
    PXE_SIG,                    /* The signature */
    0x74,                       /* length in bytes, sizeof(pxe_t) */
    0x0,                        /* checksum */
    PXE_REV,                    /* structure version */
    0,                          /* reserved, must be zero */
    &nilo_undi_romid,           /* pointer to UNDI ROMID  */
    &nilo_bc_romid,             /* pointer to base-code ROMID */
    0,                          /* 16bit API entry point */
    (void *)call_pxe_api,       /* 32bit API entry point */
    0,                          /* DHCP/TFTP callback */
    0,                          /* reserved, must be zero */
    0,                          /* # of segment descriptors */
    0,                          /* 1st segment descriptor in GDT */
    0,                          /* stack pointer */
    0,                          /* UNDI data pointer */
    0,                          /* UNDI code pointer */
    0,                          /* UNDI code size */
    0,                          /* Base Code data pointer (_edata) */
    0,                          /* Base Code code pointer (_etext) */
    0                           /* Base Code code size */
};
#endif

/* This is our default PXENV+ structure as used by all older PXE compliant
   code. Intel will be phasing out the use of this structure in future
   versions of the PXE spec. */
pxenv_t *pxenv_data_struct;
#if 1
pxenv_t nilo_pxenv = {
    PXENV_SIG,                  /* The signature */
    PXENV_REV,                  /* version */
    0x30,                       /* length in bytes, sizeof(pxenv_t) */
    0x0,                        /* checksum */
    (void *)call_pxe_api,       /* Real mode API entry point */
    0,                          /* Protected mode API entry point */
                                /* must be zero */
    0,                          /* Real mode stack segment. */
    0,                          /* stack segment size in bytes. */
    0,                          /* Real mode base-code code segment. */
    0,                          /* Base-code code segment size */
    0,                          /* Real mode base-code data segment. */
    0,                          /* Base-code data segment size */
    0,                          /* Real mode UNDI data segment. */
    0,                          /* UNDI data segment size in bytes. */
    0,                          /* UNDI code segment size in bytes. */
    0,                          /* Real mode segment:offset pointer */
    &nilo_pxe                   /* pointer to PXE Runtime structure. */
};
#endif

pxenv_ret_code_t
call_pxe_api (u_int16_t index, void *param)
{
    pxenv_ret_code_t result;
    
    //    NILO_TRACE_FUNCTION;

//    pxe_data_struct->entry32point(param);

    switch (index) {
        /* UNDI functions */
    case PXENV_START_UNDI:              /* index 0x0 */
        result = pxenv_start_undi(param);
        break;
    case PXENV_UNDI_STARTUP:            /* index 0x1 */
        result = pxenv_undi_startup(param);
        break;
    case PXENV_UNDI_CLEANUP:            /* index 0x2 */
        result = pxenv_undi_cleanup(param);
        break;
    case PXENV_UNDI_INITIALIZE:         /* index 0x3 */
        pxenv_undi_initialize(param);
        break;
    case PXENV_UNDI_RESET_NIC:      /* index 0x4 */
        result = pxenv_undi_reset_nic(param);
        break;
    case PXENV_UNDI_SHUTDOWN:           /* index 0x5 */
        result = pxenv_undi_shutdown(param);
        break;
    case PXENV_UNDI_OPEN:               /* index 0x6 */
        result = pxenv_undi_open(param);
        break;
    case PXENV_UNDI_CLOSE:              /* index 0x7 */
        result = pxenv_undi_close(param);
        break;
    case PXENV_UNDI_TRANSMIT:           /* index 0x8 */
        result = pxenv_undi_transmit(param);
        break;
    case PXENV_UNDI_SET_MCAST_ADDR:     /* index 0x9 */
        result = pxenv_undi_set_mcast_addr(param);
        break;
    case PXENV_UNDI_SET_STATION_ADDR:   /* index 0xa */
        result = pxenv_undi_set_station_addr(param);
        break;
    case PXENV_UNDI_SET_PACKET_FILTER:  /* index 0xb */
        result = pxenv_undi_set_packet_filter(param);
        break;
    case PXENV_UNDI_GET_INFORMATION:    /* index 0xc */
        result = pxenv_undi_get_information(param);
        break;
    case PXENV_UNDI_GET_STATISTICS:     /* index 0xd */
        result = pxenv_undi_get_statistics(param);
        break;
    case PXENV_UNDI_CLEAR_STATISTICS:   /* index 0xe */
        result = pxenv_undi_clear_statistics(param);
        break;
    case PXENV_UNDI_INITIATE_DIAGS:     /* index 0xf */
        result = pxenv_undi_initiate_diags(param);
        break;
    case PXENV_UNDI_FORCE_INTERRUPT:    /* index 0x10 */
        result = pxenv_undi_force_interrupt(param);
        break;
    case PXENV_UNDI_GET_MCAST_ADDR:     /* index 0x11 */
        result = pxenv_undi_get_mcast_addr(param);
        break;
    case PXENV_UNDI_GET_NIC_TYPE:       /* index 0x12 */
        result = pxenv_undi_get_nic_type(param);
        break;
    case PXENV_UNDI_ISR:                /* index 0x14 */
        result = pxenv_undi_isr(param);
        break;
    case PXENV_STOP_UNDI:               /* index 0x15 */
        result = pxenv_stop_undi(param);
        break;
    case PXENV_UNDI_GET_STATE:          /* index 0x16 */
        result = pxenv_undi_get_state(param);
        break;
        /* TFTP functions */
    case PXENV_TFTP_OPEN:               /* index 0x20 */
        result = pxenv_tftp_open(param);
        break;
    case PXENV_TFTP_CLOSE:              /* index 0x21 */
        result = pxenv_tftp_close(param);
        break;
    case PXENV_TFTP_READ:               /* index 0x22 */
        result = pxenv_tftp_read(param);
        break;
    case PXENV_TFTP_READ_FILE:          /* index 0x23 */
        result = pxenv_tftp_read_file(param);
        break;
    case PXENV_TFTP_GET_FSIZE:          /* index 0x25 */
        result = pxenv_tftp_get_fsize(param);
        break;   
        /* UDP functions */
    case PXENV_UDP_OPEN:                /* index 0x30 */
        result = pxenv_udp_open(param);
        break;
    case PXENV_UDP_CLOSE:               /* index 0x31 */
        result = pxenv_udp_close(param);
        break;
    case PXENV_UDP_READ:                /* index 0x32 */
        result = pxenv_udp_read(param);
        break;
    case PXENV_UDP_WRITE:               /* index 0x33 */
        result = pxenv_udp_write(param);
        break;
    case PXENV_UNLOAD_STACK:            /* index 0x70 */
        pxenv_unload_stack(param);
        break;
    case PXENV_GET_BINL_INFO:
        pxenv_get_cached_info(param);
        break;
    case PXENV_RESTART_TFTP:		 /* index 0x73 */
        pxenv_restart_tftp(param);
        break;
    case PXENV_START_BASE:              /*  index 0x75 */
        pxenv_start_base(param);
        break;
    case PXENV_STOP_BASE:               /*  index 0x76 */
        pxenv_stop_base(param);
        break;
    default:
        return (PXENV_EXIT_FAILURE);
    }
    
    //    NILO_TRACE_FUNCTION_RETURN;
    return (result);
}

/* We find the pxe and pxenv structures by scanning memory between the
 * top of the free memory at boot time, and 0x0A0000 (640k). If the top
 * of memory can't be found, start from 0x0A0000 and scan down towards
 * 0x10000 (64k). It's usually found between 0x80000 (512k) and
 * 0x0A0000. It can be found for looking for the signature, and then
 * checking the version number and the checksum. */
void *
find_pxe_data (pxe_type_t type)
{
    u_int8_t *i, size, *sig, rev;
    u_int8_t *upper = (unsigned char *)UPPER_PXE_MEM;
    u_int8_t *lower = (unsigned char *)LOWER_PXE_MEM;

    NILO_TRACE_FUNCTION;

    /* Depending on which data structure we're looking for, adjust
       the search criteria. */
    if (type == PXE) {
        size = 4;
        sig = PXE_SIG;
        rev = PXE_REV;
    } else {
        size = 6;
        sig = PXENV_SIG;
        rev = PXENV_REV;
    }

    /* FIXME: for now, force the address to "just work". This is only
       here so find_pxe() can actually find something this is just for
       testing purposes. */
#if 1
    lower = (unsigned char *)&_etext;
    upper = (unsigned char *)&_end;
#else
    lower = (unsigned char *)LOWER_PXE_MEM;
    upper = (unsigned char *)UPPER_PXE_MEM;
#endif

    /* Scan backwards through memory looking for the signature */

    for (i=upper; i>lower;i--) {
        if (memcmp (i, sig, size) == 0) {
            /* We found it. Now see if it's usable. */

            /* FIXME: We don't want to find our own data */
//            if (i == &nilo_pxe && i == &nilo_pxenv)
//                NILO_TRACE_MSG ("Oops, found our internal structure\r\n");

            /* Look for the version number. As the names of this field 
               are different for the PXE and PXENV structures, we use an
               offset of 6 bytes, which is the same for both structures. */
            if (*(i+6) != rev) {
                NILO_TRACE_MSG ("PXE version 0x02x too low.\r\n", *(i+6));
            } else {
                NILO_TRACE_MSG ("Found the %s data structure at 0x%04x\r\n",
                                (type == PXE) ? "!PXE" : "PXENV+", i);
                NILO_TRACE_FUNCTION_RETURN;
                return (i);
            }
            /* Check the checksums */
        }
    }

    /* Since we couldn't find the data structure, use our internal one */
    if (i == lower) {
        NILO_TRACE_MSG ("Didn't find the data\r\n");
        NILO_TRACE_FUNCTION_RETURN;
//        if (type == PXE) {
//            return (&nilo_pxe);
//        } else {
//            return (&nilo_pxenv);
//        }
    }

    NILO_TRACE_FUNCTION_RETURN;
    return (0);
}

